#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "calc_energy.h"
#include "pfunc.h"


/* unused */
int do_shrink;
int constraint;
int cut_point = -1;

void verify_struc(const char *struc);
void struc2array(const char *struc);
void usage();


int main(int argc, char **argv)
{
    int i;
    char *struc = NULL;
    double energy;              /* the total energy of the given structure */
    char **seqs_r, **seqs_s;      /* the sequence alignments of R and S */
    char *input_nameR = "./data/ralign.in";
    char *input_nameS = "./data/salign.in";
    
    for (i=1; i<argc; i++) {
        if (argv[i][0] == '-') {
            switch (argv[i][1]) {
                case 'i':
                    if(i==argc-1 || i==argc-2 || argv[i+1][0]=='-')
                        usage();
                    
                    input_nameR = argv[++i];
                    input_nameS = argv[++i];
                    break;
                case 'T':
                    if (i == argc-1)
                        usage();
                    
                    if (sscanf(argv[++i], "%lf", &temperature) == 0)
                        usage();
                    break;
                case 'h':
                    usage();
                    break;
                default:
                    break;
            }
        }
    }
    
    read_align(&seqs_r, &seqs_s, input_nameR, input_nameS);
    initialize_array();
    encode_align(seqs_r, seqs_s);
    
    
    struc = (char *) space(sizeof(char)*(len_r+len_s+3));
    fprintf(stdout, "Please input a structure of your sequences:\n");
    fscanf(stdin, "%s", struc);
    verify_struc(struc);

    /* convert the structure into integer representation */
    struc2array(struc);

    /* calculate the energy of the structure */
    energy = calc_rip_energy(seqs_r, seqs_s);

    fprintf(stderr, "The energy of your structure is: %f kcal/mol\n", energy/100);
    
    free_array();
    free(struc);
    
    return EXIT_SUCCESS;
}


void verify_struc(const char *struc)
{
    int i;
    int t1=0, t2=0;
    int n_separator = 0;              /* the vertical separator bar */

    if (struc[len_r] != '|') {
        fprintf(stderr, "%s,%d: error: there is no \'|\' in the structure.\n", __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    }
    if (strlen(struc) != (len_r+len_s+1)) {
        fprintf(stderr, "%s,%d: error: the length of the strucutre is incorrect.\n", __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    }

    /* valid characters */
    for (i=0; i<strlen(struc); i++) {
        if (struc[i]!='(' && struc[i]!=')' && struc[i]!='[' && struc[i]!=']'
            && struc[i]!='.' && struc[i]!='|') {
            fprintf(stderr, "%s,%d: error: there exist some invalid characters in the structure.\n", __FILE__, __LINE__);
            exit(EXIT_FAILURE);
        }
    }

    /* the matching of bracket */
    for (i=0; i<strlen(struc); i++) {
        switch(struc[i]) {
            case '|':
                n_separator++;
                break;
            case '(':
                t1++;
                break;
            case ')':
                t1--;
                break;
            case '[':
                t2++;
                break;
            case ']':
                t2--;
                break;
            default:
                break;
        }
    }
    if (n_separator > 1) {
        fprintf(stderr, "%s,%d: error: there exist one more vertical separator in the structure.\n", __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    }
    if (t1 != 0) {
        fprintf(stderr, "%s,%d: error: the brackets of secondary structure are not matching.\n", __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    }
    if (t2 != 0) {
        fprintf(stderr, "%s,%d: error: the brackets of interaction are not matching.\n", __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    }

    return;                     /* an invalid structure */
}

void struc2array(const char *struc)
{
    int i, l, shift = 0;
	int stack1[500], index1, stack2[500], index2;
	int *pair, m;

	l = strlen(struc);

	pair = (int *) space(sizeof(int)*(l+1));
	pair[0] = l;
    
	index1 = 0;
    index2 = 0;
	for (i=0; i<l; i++) {
        if (struc[i] == '|') {
            shift = 1;
            m = i;
        }
        if (struc[i] == '(') {
            stack1[index1] = i-shift+1;
            index1++;
        }
        else if (struc[i] == ')') {
            pair[i-shift+1] = stack1[index1-1];
            pair[stack1[index1-1]] = i-shift+1;
            index1--;
        }
        else if (struc[i] == '[') {
            stack2[index2] = i-shift+1;
            index2++;
        }
        else if (struc[i] == ']') {
            pair[i-shift+1] = stack2[index2-1];
            pair[stack2[index2-1]] = i-shift+1;
            index2--;
        }
	}

    if (index1!=0 || index2!=0) {
        fprintf(stderr, "%s,%d: error: stack is not empty!\n", __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    }

    /*
    for (i=0; i<(l+1); i++)
        fprintf(stdout, "%d  ", pair[i]);
    fprintf(stdout, "\n");
    */

    for (i=1; i<=len_r; i++) {
        if (pair[i] == 0)
            strand_r[i] = SR;
        else if (pair[i] <= len_r)
            strand_r[i] = SR;
        else if (pair[i] > len_r)
            strand_r[i] = SS;
    }
    for (i=1; i<=len_s; i++) {
        if (pair[i+len_r] == 0)
            strand_s[i] = SS;
        else if (pair[i+len_r] <= len_r)
            strand_s[i] = SR;
        else if (pair[i+len_r] > len_r)
            strand_s[i] = SS;
    }
    strand_r[0] = strand_s[0] = -1;
    
    for (i=1; i<=len_r; i++) {
        if (strand_r[i] == SR)
            int_r[i] = pair[i];
        else if (strand_r[i] == SS)
            int_r[i] = pair[i]-len_r;
    }
    for (i=1; i<=len_s; i++) {
        if (strand_s[i] == SS)
            int_s[i] = (pair[i+len_r]==0) ? 0 : (pair[i+len_r]-len_r);
        else if (strand_s[i] == SR)
            int_s[i] = pair[i+len_r];
    }
    
    free(pair);
}

void usage()
{
    fprintf(stdout,
            "usage: RNAengy [-i ralign salign] [-T temperature] [-h]\n"
            "\n"
            "options:\n"
            "\t -i ralign salign  read sequences from the two files. The default are\n"
            "\t                     \"./data/align.u4.in\" and \"./data/align.u6.in\".\n"
            "\t -T temperature    calculate energy at the given temperature.\n"
            "\t -h                show help.\n");

    exit(EXIT_FAILURE);
}
