#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include "pfunc.h"
#include "sample.h"
#include "prob.h"

int cut_point = -1;
int sample;
long num_sample;
int single_sample;
int top_frequency;
int all_sample;
int detail;
int constraint;
int do_prob;
int do_ps_bpm;      //generate ps file of base pairing probability matrix
int do_profile;
int do_shrink;
int n_top;        //number of top hybrids

struct plist *pre, *mfe, *profile, *empty;

void usage(void);

int main(int argc, char **argv)
{
    int i, j, r;
    FILE *fp_output;
    double lr, ls;
    double mem_int, mem_double;
    double mem_pf, mem_bp = 0.0, mem_total;
    char **seqs_r = NULL;
    char **seqs_s = NULL;
    char **seqs = NULL, *ptr;
    char *cons_string = NULL;
    char *input_file_nameR = "./data/ralign.in";
    char *input_file_nameS = "./data/salign.in";
    char *output_file_name = "./result/align.out";
    char *ps_file_name = "./result/dot.ps";
    char *pro_file_name = "./result/profile.ps";
    char *cons_file_name = "./data/rip.cstr";

    srand48(time(NULL));
    
    /* default parameters */
    sample = 0;
    single_sample = 0;
    num_sample = 10000;
    top_frequency = 20;
    all_sample = 0;
    do_prob = 0;
    do_ps_bpm = 0;
    do_profile = 0;
    detail = 0;
    constraint = 0;
    do_shrink = 0;
    n_top = 5;

    /* ------------------- process command line parameters ------------------- */
    if (argc == 1)
        printf("The default is to calculate partition function only, and the result\n"
               "will be in \"%s\"\n\n", output_file_name);
    for (i=1; i<argc; i++) {
        if (argv[i][0]=='-')
            switch (argv[i][1]) {
                case 'S':
                    sample = 1;
                    break;
                case 'P':
                    do_prob = 1;
                    break;
                default:
                    break;
            }
    }
    for (i=1; i<argc; i++) {
        if (argv[i][0]=='-')
            switch (argv[i][1])
            {
                case 'r':
                    do_shrink = 1;
                    break;
                case 'i':                       //input file name
                    if(i==argc-1 || i==argc-2 || argv[i+1][0]=='-')
                        usage();

                    input_file_nameR = argv[++i];
                    input_file_nameS = argv[++i];
                    break;
                case 'o':                       //output file name
                    if(i==argc-1 || argv[i+1][0]=='-')
                        usage();

                    output_file_name = argv[++i];
                    break;
                case 'f':                       //postscript file name
                    if (!do_prob)
                        usage();
                    
                    do_ps_bpm = 1;
                    if(i<argc-1 && argv[i+1][0]!='-')
                        ps_file_name = argv[++i];
                    break; 
                case 'p':                       //output profile
                    if (!do_prob)
                        usage();
                    
                    do_profile = 1;
                    if (i<argc-1 && argv[i+1][0]!='-')
                        pro_file_name = argv[++i];
                    break;
                case 'n':                       //number of sampling structures
                    if (!sample || i==argc-1)
                        usage();                    

                    r = sscanf(argv[++i], "%ld", &num_sample);
                    if (r == 0)
                        usage();
                    break;
                case 'S':
                    sample = 1;
                    break;
                case 's':                       //sample a single structure
                    if (!sample)
                        single_sample = 1;
                    else
                        fprintf(stderr, "Warning: -s option is ignored!\n\n");
                    break;
                case 'a':                       //show all samples in @output_file_name
                    if (!sample)
                        usage();
                    
                    all_sample = 1;
                    break;
                case 'P':                       //calculate base pair probabilities
                    do_prob = 1;
                    break;
                case 'C':                       //with constraints
                    constraint = 1;

                    if (i<argc-1 && argv[i+1][0]!='-')
                        cons_file_name = argv[++i];
                    break;
                case 'd':                       //show details of base pair probabilities, profiles and hybrid blocks
                    if (!do_prob)
                        usage();
                    
                    detail = 1;
                    break;
                case 't':                       //top frequency
                    if (!sample || i==argc-1)
                        usage();
                    
                    r = sscanf(argv[++i], "%d", &top_frequency);
                    if (r == 0)
                        usage();
                    break;
                case 'T':                       //top hybrid
                    if (!sample || i==argc-1)
                        usage();

                    r = sscanf(argv[++i], "%d", &n_top);
                    if (r == 0)
                        usage();
                    break;
				/*	
                case 'F':
                    if (!sample || i==argc-1)
                        usage();

                    r = sscanf(argv[++i], "%lf", &cv_fact);
                    if (r == 0)
                        usage();
                    break;
				*/
                case 'h':                       //show help
                    usage();
                    break;
                default:
                    usage();
            }
    }   //end for (i=1; i<argc; i++)
    
    mfe = (struct plist *) space(MAXPRO*sizeof(struct plist));
    pre = (struct plist *) space(MAXPRO*sizeof(struct plist));
    profile = (struct plist *) space(MAXPRO*sizeof(struct plist));
    empty = (struct plist *)space(MAXPRO*sizeof(struct plist));
    
    fp_output = fopen(output_file_name, "w");
    if (!fp_output) {
        printf("Can not open output file %s\n", output_file_name);
        exit(EXIT_FAILURE);
    }
    
    /* initialization */
    read_align(&seqs_r, &seqs_s, input_file_nameR, input_file_nameS);
    if (constraint)
        cons_string = read_align_constraint(cons_file_name);
    initialize_params();
    initialize_matrix();

    /* -------------- calculate used memories ---------------- */
    //memories of partition function
    lr = (double) len_r;
    ls = (double) len_s;
    mem_int = (lr+ls+2) + (num_align+1)*(lr+ls+4)
        + (num_align+2)*(lr*(lr+1)/2+ls*(ls+1)/2+(lr+1)*(ls+2)+7);
    mem_int *= sizeof(int);
    mem_double = lr*(lr+1)/2+ls*(ls+1)/2+(lr+1)*(ls+2)+2
        + 5*(lr*(lr+1)/2+ls*(ls+1)/2+4) + 62*(lr*(lr+1)/2+2)*(ls*(ls+1)/2+3);
    mem_double *= sizeof(double);
    mem_pf = mem_int+mem_double;
	printf("Memories for partition function:  %lf(KB) = %lf(MB) = %lf(GB)\n",
           mem_pf/((double) 1024), mem_pf/((double) (1024*1024)), mem_pf/((double) (1024*1024*1024)));
    //memories of base pairing probabilities
    if (do_prob) {
        mem_bp = 6*(lr*(lr+1)/2+ls*(ls+1)/2+4) + (lr+1)*(ls+2) + 66*(lr*(lr+1)/2+2)*(ls*(ls+1)+3);
        mem_bp *= sizeof(double);
        printf("Memories for base pairing probabilities:  %lf(KB) = %lf(MB) = %lf(GB)\n",
               mem_bp/((double) 1024), mem_bp/((double) (1024*1024)), mem_bp/((double) (1024*1024*1024)));
    }
    //total memories
    mem_total = mem_pf+mem_bp;
    printf("Total used memories:  %lf(KB) = %lf(MB) = %lf(GB)\n\n",
           mem_total/((double) 1024), mem_total/((double) (1024*1024)), mem_total/((double) (1024*1024*1024)));

    /* --------------------------- calculate partition function ------------------------ */
    printf("Calculating alignments partition function... ");
    fflush(stdout);
    calc_align_pfunc(fp_output, seqs_r, seqs_s, cons_string);
    printf("done\n");

    /* --------------------------- sample structures ----------------------------- */
    if (sample) {
        int result;
        
        printf("Sampling alignment structures... ");
        fflush(stdout);
        result = do_sample(fp_output, seqs_r, seqs_s, num_sample, all_sample, top_frequency, mfe);
        printf("done\n");
        if (result != 0)
            fprintf(stderr, "***Could not enforce all constraints***\n");
    }
    else if (single_sample) {
        printf("Sampling single structures... ");
        fflush(stdout);
        do_single_sample(fp_output, seqs_r, seqs_s);
        printf("done\n");
    }

    /* -------------------- calculate base pair probabilities ------------------- */
    if (do_prob) {
        initialize_prob();
        printf("Calculating base pairing probabilities... ");
        fflush(stdout);
        calc_bp_prob(fp_output, detail, pre);
        printf("done\n");

        //join two alignments together
        seqs = (char **) space(sizeof(char *)*(num_align+1));
        ptr = (char *) space(sizeof(char)*(num_align+1)*(len_r+len_s+4));
        for (i=0; i<(num_align+1); i++) {
            seqs[i] = ptr;
            ptr += (len_r+len_s+4);
        }
        for (i=1; i<=num_align; i++) {
            for (j=0; j<len_r; j++)
                seqs[i][j] = seqs_r[i][j];
            for (j=len_r; j<(len_r+len_s+4); j++)
                seqs[i][j] = seqs_s[i][j-len_r];
        }

        cut_point = len_r+1;

        if (do_ps_bpm)
            (void) PS_dot_plot_list(seqs[1], ps_file_name, pre, mfe, "doof");
        
        /* generate probability profile */
        if (do_profile) {
            printf("Generating probability profile... ");
            fflush(stdout);
            generate_prob_profile(fp_output, detail, profile);
            (void) PS_dot_plot_list(seqs[1], pro_file_name, profile, empty, "doof");
            printf("done\n");
        }

        free(seqs[0]); free(seqs);
        free_prob();
    }

    fclose(fp_output);
    free(pre); free(mfe); free(profile); free(empty);
    free_matrix();
    free(cons_string);
    free(seqs_r[0]); free(seqs_r);
    free(seqs_s[0]); free(seqs_s);
    if (constraint)
        free(BP);

    return 0;
}

void usage(void)
{
    printf("Usage:  RNArip [-i rfile sfile] [-o outputfile] [-r] [-C [cstrfile]] [-s] [-S OPTIONS1] [-P OPTIONS2] [-h]\n"
		   "\'RNArip\' calculates partition function of seqence alignment, samples random\n"
           "structures and calculates base pairing probabilities.\n\n"
		   "Examples:\n"
		   " RNArip                    #Only calculate partition function of seqence alignments\n"
		   " RNArip -S -n 1000 -P -d   #Sample 1000 random structures and calculate base\n"
		   "                               pairing probabilities in detail\n\n"
           " Main operation:\n\n"
		   "	-i rfile sfile      specify the input file(FASTA format), and both are from\n" 			   "                              5\' direction. The default are \"./data/ralign.in\" and\n"
           "                              \"./data/salign.in\"\n"
           "	-o outputfile       specify the output file.\n"
           "                              The default is \"./result/align.out\"\n"
		   "	-r                  shrink alignments.\n"
		   "	-C [cstrfile]       calculate partition function subject to constraints.\n"
           "                              The default constraint file is \"./data/rip.cstr\"\n"
		   "	-s                  sample single random structure\n"
		   "	-S                  sample random structures\n"
           "	-P                  calculate base pairing probabilities\n\n"
		   " OPTIONS1:\n\n"
		   "	-a                  show all samples\n"
		   "	-t topfrequency     specify the number of top frequency.\n"
           "                              The default is 20\n"
           "	-n numsample        sample numsample random samples.\n"
           "                              The default is 10000\n\n"
		   " OPTIONS2:\n\n"
           "	-d                  show details about the calculation. (base pairing\n"
           "                              probability, base pairing frequency and profiles)\n"
           "	-f [psfile]         specify the postscript file of base pairing\n"
           "                              probability matrix. The default is \"./result/dot.ps\"\n"
           "	-p [profile]        calculate profile, and specify its postscript file.\n"
           "                              The default is \"./result/profile.ps\"\n"
           "	-h                  show help.\n");
    
    exit(0);
}
