#include <stdio.h>
#include <stdlib.h>
#include "inc.h"
#include "energy.h"


paramT *P;


void read_params()
{
    P = scale_parameters();
}

inline int calc_hairpin_energy(const char *string, int size, int type, int si1, int sj1)
{
  int energy;
  
  energy = (size <= 30) ? P->hairpin[size] :
      P->hairpin[30]+(int)(P->lxc*log((size)/30.));
  
  if (tetra_loop)
      if (size == 4) { /* check for tetraloop bonus */
          char tl[7] = {0}, *ts;
          
          strncpy(tl, string, 6);
          if ((ts=strstr(P->Tetraloops, tl)))
              energy += P->TETRA_ENERGY[(ts - P->Tetraloops)/7];
      }
  
  if (size == 3) {
      char tl[6] = {0}, *ts;
      
      strncpy(tl, string, 5);
      if ((ts=strstr(P->Triloops, tl)))
          energy += P->Triloop_E[(ts - P->Triloops)/6];
      
      if (type > 2)  /* neither CG nor GC */
          energy += P->TerminalAU; /* penalty for closing AU GU pair */
  }
  else  /* no mismatches for tri-loops */
      energy += P->mismatchH[type][si1][sj1];
  
  return energy;
}

inline int calc_interior_energy(int n1, int n2, int type, int type_2,
                                int si1, int sj1, int sp1, int sq1)
{
  /* compute energy of degree 2 loop (stack bulge or interior) */
  int nl, ns, energy;

  if (n1 > n2) {
      nl = n1;
      ns = n2;
  }
  else {
      nl = n2;
      ns = n1;
  }
  
  if (nl == 0)
      return P->stack[type][type_2];    /* stack */
  
  if (ns==0) {                       /* bulge */
      energy = (nl<=MAXLOOP) ? P->bulge[nl] :
          (P->bulge[30]+(int)(P->lxc*log(nl/30.)));
      
      if (nl == 1)
          energy += P->stack[type][type_2];
      else {
          if (type > 2)
              energy += P->TerminalAU;
          if (type_2 > 2)
              energy += P->TerminalAU;
      }
      
      return energy;
  }
  else {                             /* interior loop */
      if (ns == 1) {
          if (nl == 1)                      /* 1x1 loop */
              return P->int11[type][type_2][si1][sj1];
          
          if (nl == 2) {                   /* 2x1 loop */
              if (n1 == 1)
                  energy = P->int21[type][type_2][si1][sq1][sj1];
              else
                  energy = P->int21[type_2][type][sq1][si1][sp1];
              
              return energy;
          }
      }
      else if (n1==2 && n2==2)         /* 2x2 loop */
          return P->int22[type][type_2][si1][sp1][sq1][sj1];

      //generic interior loop (no else here!)
      energy = (n1+n2<=MAXLOOP) ? (P->internal_loop[n1+n2]) :
          (P->internal_loop[30]+(int)(P->lxc*log((n1+n2)/30.)));
      
      energy += MIN(MAX_NINIO, (nl-ns)*P->F_ninio[2]);
      
      energy += P->mismatchI[type][si1][sj1]+P->mismatchI[type_2][sq1][sp1];
  }
  
  return energy;
}
