#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "eps.h"


void generate_eps(int n_seq, char **seqs_r, char **seqs_s, const char *name_r, const char *name_s)
{
    FILE *ptr_result;
    char ch;
    int len_r, len_s;
    int i, j, *tmp_s, len_hyb = 0;

    ptr_result = fopen("./rip.eps", "w");
    if (!ptr_result) {
        fprintf(stderr, "Error top open %s to write!\n", "./rip.eps");
        exit(EXIT_FAILURE);
    }
    
    len_r = strlen(seqs_r[1]);
    len_s = strlen(seqs_s[1]);

    for (i=1; i<=n_seq; i++) {
        for(j=0; j<len_s/2; j++) {      //inverse the @seqs_s sequence
            ch = seqs_s[i][j];
            seqs_s[i][j] = seqs_s[i][len_s-j-1];
            seqs_s[i][len_s-j-1] = ch;
        }
    }
    
    tmp_s = (int *) malloc(sizeof(int)*(len_s+1));
	memset(tmp_s, 0, sizeof(int)*(len_s+1));
    tmp_s[0] = len_s;
	for(i=1; i<=len_s; i++) {        //exchange the relation array of @s
		if(s[i]!=0 && i<s[i]) {
			tmp_s[len_s-i+1] = len_s-s[i]+1;
            tmp_s[len_s-s[i]+1] = len_s-i+1;
        }
	}

    /* ------------------------------ write eps file --------------------------------- */
	//write the head infomation of eps file
	fprintf(ptr_result,
            "%%!PS-Adobe-3.0 EPSF-3.0\n"
            "%%%%BoundingBox: 0 0 300 480\n"
            "%%%%LanguageLevel: 3\n"
            "%%%%Creator: rip\n"
            "%%%%EndComments\n\n");

    fprintf(ptr_result, "%%------------------------ data ----------------------------\n");
    //write array @r
    fprintf(ptr_result, "/arr [\n\t[");
    for(i=0; i<=r[0]; i++)
        fprintf(ptr_result, "%d ", r[i]);
    fprintf(ptr_result, "]\n\t[");
    
    //write array @s
    for(i=0; i<=len_s; i++)
        fprintf(ptr_result, "%d ", tmp_s[i]);
    fprintf(ptr_result, "]\n] def\n");
    free(tmp_s);
    
    //write array @hybrid
    fprintf(ptr_result, "/inc [\n");
    for (i=1; i<=hybrid[0]; i++) {
        if (hybrid[i] != 0) {
            len_hyb++;
            fprintf(ptr_result, "\t[%d %d]\n", i, hybrid[i]);
        }
    }
    fprintf(ptr_result, "] def\n\n");

    //write name of sequences
    fprintf(ptr_result, "/nameR (%s) def\n", name_r);
    fprintf(ptr_result, "/nameS (%s) def\n", name_s);
    
    //write sequences
    fprintf(ptr_result, "/seqR [\n");
    for(i=1; i<=n_seq; i++) {
        fprintf(ptr_result, "\t(");
        for (j=0; j<strlen(seqs_r[1]); j++)
            fprintf(ptr_result, "%c", seqs_r[i][j]);
        fprintf(ptr_result, ")\n");
    }
    fprintf(ptr_result, "] def\n");
    fprintf(ptr_result, "/seqS [\n");
    for(i=1; i<=n_seq; i++) {
        fprintf(ptr_result, "\t(");
        for (j=0; j<strlen(seqs_s[1]); j++)
            fprintf(ptr_result, "%c", seqs_s[i][j]);
        fprintf(ptr_result, ")\n");
    }
    fprintf(ptr_result, "] def\n");

    //write hybrid regions
    fprintf(ptr_result, "/region [\n");
    for (i=0; i<n; i++)
        fprintf(ptr_result, "\t[%d %d (%.1f%%)]\n", hyb[i].i, hyb[i].j, hyb[i].pro*100);
    fprintf(ptr_result, "] def\n\n");

    fprintf(ptr_result, "%%----------------------- variables & procedures -------------------------\n");
	//some variables and procedures
    fprintf(ptr_result, "/width 300 def\n"
            "/height 480 def\n"
            "/LM 20 def\n"
            "/RM 20 def\n"
            "/TM 20 def\n"
            "/BM 20 def\n");
	fprintf(ptr_result, "/len_r %d def\n", len_r);
    fprintf(ptr_result, "/len_s %d def\n", len_s);
	fprintf(ptr_result, "/len_hyb %d def\n", len_hyb);
    fprintf(ptr_result, "/n_seq %d def\n", n_seq);
	fprintf(ptr_result, "/fontsize 4 def\n"
            "/lineWidth 40 def\n"
            "/xscale 1 def\n"
            "/yscale .5 def\n\n"
            "/drawSeqR {\n"
            "    .4 setgray\n\n"
            "    /i 1 def\n"
            "    1 1 n_seq {\n"
            "        0 lineWidth 2 div i 1 sub fontsize mul add moveto\n"
            "        seqR n_seq i sub get show\n\n"
            "        /i i 1 add def\n"
            "    } for\n"
            "} def\n"
            "/drawSeqS {\n"
            "    .4 setgray\n\n"
            "    /i 1 def\n"
            "    1 1 n_seq {\n"
            "        0 lineWidth 2 div neg i fontsize mul sub moveto\n"
            "        seqS i 1 sub get show\n\n"
            "        /i i 1 add def\n"
            "    } for\n"
            "} def\n");

    /* -------------------- draw ruler of sequences -------------------- */
    /* ruler R */
    fprintf(ptr_result,
            "/ruler_r {\n"
            "    /str 10 string def\n"
            "    .4 setgray\n\n"
            "    0 lineWidth 2 div fontsize sub moveto\n"
            "    (");
    for (i=1; i<=len_r; i++)
        fprintf(ptr_result, "%d", i%10);
    fprintf(ptr_result, ") show\n\n");
    fprintf(ptr_result,
            "    %%draw vertical bar\n"
            "    gsave\n"
            "    0 lineWidth 2 div fontsize 1.5 mul sub translate\n"
            "    1 .5 scale\n"
            "    1 1 len_r {\n"
            "        /i exch def\n"
            "        /tmp i 10 mod def\n"
            "        tmp 0 eq {\n"
			"            i 1 sub (0) stringwidth pop mul 0 moveto\n"
			"            (|) show } if\n"
            "    } for\n"
            "    grestore\n\n"
            "    %%draw scale\n"
            "    gsave\n"
            "    0 lineWidth 2 div fontsize 2.5 mul sub translate\n"
            "    1 1 len_r {\n"
            "        /i exch def\n"
            "        /tmp i 10 mod def\n"
            "        tmp 0 eq {\n"
			"            i 1.5 sub (0) stringwidth pop mul 0 moveto\n"
			"            i str cvs show } if\n"
            "    } for\n"
            "    grestore\n"
            "} def\n");
    /* ruler S */
    fprintf(ptr_result,
            "/ruler_s {\n"
            "    /str 10 string def\n"
            "    .4 setgray\n\n"
            "    0 lineWidth 2 div neg moveto\n"
            "    (");
    for (i=1; i<=len_s; i++)
        fprintf(ptr_result, "%d", i%10);
    fprintf(ptr_result, ") show\n\n");
    fprintf(ptr_result,
            "    %%draw vertical bar\n"
            "    gsave\n"
            "    0 lineWidth 2 div neg fontsize add translate\n"
            "    1 .5 scale\n"
            "    1 1 len_s {\n"
            "        /i exch def\n"
            "        /tmp i 10 mod def\n"
            "        tmp 0 eq {\n"
			"            i 1 sub (0) stringwidth pop mul 0 moveto\n"
			"            (|) show } if\n"
            "    } for\n"
            "    grestore\n\n"
            "    %%draw scale\n"
            "    gsave\n"
            "    0 lineWidth 2 div neg fontsize 1.5 mul add translate\n"
            "    1 1 len_s {\n"
            "        /i exch def\n"
            "        /tmp i 10 mod def\n"
            "        tmp 0 eq {\n"
			"            i 1.5 sub (0) stringwidth pop mul 0 moveto\n"
			"            i str cvs show } if\n"
            "    } for\n"
            "    grestore\n"
            "} def\n");

    fprintf(ptr_result,
            "/drawArcR {\n"
            "    .1 setgray\n\n"
            "    /tmp arr 0 get def\n"
            "    /i 1 def \n"
            "    1 1 len_r {\n"
            "        tmp i get\n"
            "        lt {\n"
            "            /x tmp i get i add 1 sub step mul 2 div def\n"
            "            /y lineWidth 2 div n_seq 1 sub fontsize mul fontHeight add add def\n"
			"            /r tmp i get i sub step mul 2 div def\n"
			"            gsave\n"
			"            x y translate\n"
			"            xscale yscale scale\n"
			"            0 0 r 0 180 arc\n"
			"            stroke\n"
			"            grestore\n"
            "        }if\n"
            "        /i i 1 add def\n"
            "    }for\n"
            "} def\n"
            "/drawArcS {\n"
            "    .1 setgray\n\n"
            "    /tmp arr 1 get def\n"
            "    /i 1 def\n"
            "    1 1 len_s {\n"
            "        tmp i get\n"
            "        lt {\n"
            "            /x tmp i get i add 1 sub step mul 2 div def\n"
            "            /y lineWidth neg 2 div n_seq fontsize mul sub fontsize fontHeight sub 2 div sub def\n"
            "            /r tmp i get i sub step mul 2 div def\n"
            "            gsave\n"
            "            x y translate\n"
            "            xscale yscale scale\n"
            "            0 0 r 180 360 arc\n"
            "            stroke\n"
            "            grestore\n"
            "         }if\n"
            "         /i i 1 add def\n"
            "   }for\n"
            "} def\n"
            "/drawHybrid {\n"
            "    .1 setgray\n\n"
            "    /i 0 def\n"
            "    1 1 len_hyb {\n"
            "        /tmp inc i get def\n\n"             
            "        /xr tmp 0 get 0.5 sub step mul def\n"
            "        /yr lineWidth fontsize sub fontHeight add 2 div def\n"
            "        /xs tmp 1 get 0.5 sub step mul def\n"
            "        /ys lineWidth 2 div neg fontsize sub fontHeight add def\n"
            "        xr yr moveto\n"
            "        xs ys lineto\n"
            "        stroke\n\n"
            "        /i i 1 add def\n"
            "    }for\n"
            "} def\n\n"
            "/maxHeightR {\n"
            "	/i 1 def\n"
            "	/maxR 1 def\n"
            "	/tmp arr 0 get def\n\n"
            "   1 1 len_r {\n"
            "		tmp i get lt {\n"
            "			maxR tmp i get i sub lt {\n"
            "				/maxR tmp i get i sub def\n"
            "			} if\n"
            "		} if\n"
            
            "		/i i 1 add def\n"
            "	} for\n\n"
            "	/maxR maxR step mul 2 div yscale mul def\n"
            "} def\n"
            "/maxHeightS {\n"
            "	/i 1 def\n"
            "	/maxS 1 def\n"
            "	/tmp arr 1 get def\n\n"
            "	1 1 len_s {\n"
            "		tmp i get lt {\n"
            "			maxS tmp i get i sub lt {\n"
            "			/maxS tmp i get i sub def\n"
            "			} if\n"
            "		} if\n"
            
            "		/i i 1 add def\n"
            "	} for\n\n"
            "	/maxS maxS step mul 2 div yscale mul def\n"
            "} def\n\n"
            "/drawRegion {\n"
            "    region {\n"
            "	     aload pop\n"
            "		 3 1 roll exch\n"
            "		/st exch def\n"
            "		/nd exch def\n"
            "		/prob exch def\n"
            "		/tmp arr 0 get def\n"
            "\n"
            "		gsave\n"
            "		0 lineWidth 2 div fontsize n_seq mul add fontsize fontHeight sub sub translate\n"
            "		0.7 0.9 0.8 setrgbcolor\n"
            "\n"
            "		newpath\n"
            "		st 1 sub step mul 0 moveto\n"
            "		nd step mul 0 lineto\n"
            "		nd step mul fontsize n_seq mul neg fontsize fontHeight sub add lineto\n"
            "		st 1 sub step mul fontsize n_seq mul neg fontsize fontHeight sub add lineto\n"
            "		closepath\n"
            "		fill\n"
            "\n"
            "		0 0 0 setrgbcolor\n"
            "		/Helvetica-Bold findfont 6 scalefont setfont\n"
            "		.1 setgray\n"
            "		st nd add step mul 2 div (99.9%%) stringwidth pop 2 div sub fontsize fontHeight sub moveto\n"
            "		prob show\n"
            "		grestore\n"
            "	} forall\n"
            "} def\n");
	
    fprintf(ptr_result, "%%------------------------- draw figures -------------------------\n");
    fprintf(ptr_result,
            "/Courier findfont fontsize scalefont setfont\n"
            "/dirLen (---5\'-) stringwidth pop def\n\n"
            "gsave\n"
            "/Helvetica-Bold findfont 8 scalefont setfont\n"
            "/lenNameR nameR stringwidth pop def\n"
            "/lenNameS nameS stringwidth pop def\n"
            "lenNameR lenNameS gt { /lenName lenNameR def } { /lenName lenNameS def } ifelse\n"
            "grestore\n\n"
            "/sizeR seqR 0 get stringwidth pop def\n"
            "/sizeS seqS 0 get stringwidth pop def\n"
            "sizeR sizeS gt { /lenSeq sizeR def } { /lenSeq sizeS def } ifelse\n\n"
            "/step sizeR len_r div def\n"
            "maxHeightR\n"
            "maxHeightS\n\n"
            "/fontHeight fontsize 0.72 mul def	%%the font height is around 72%% of font size\n"
            ".08 setlinewidth\n\n"
            "len_r len_s gt {/len len_r def} {/len len_s def} ifelse\n\n"
            "/oriX LM lenName add dirLen add def\n"
            "/oriY width 2 div def\n\n"
            "-90 rotate\n"
            "height neg 0 translate\n"
            "oriX oriY translate\n\n"
            "%%label regions\n"
            "drawRegion\n\n"
            "%%draw sequences\n"
            "gsave\n"
            "drawSeqR\n"
            "drawSeqS\n"
            "grestore\n\n"
            "%%draw directions of sequences\n"
            "gsave\n"
            "dirLen 0.5 mul neg lineWidth 2 div n_seq 1 sub fontsize mul add moveto\n"
            "(5\'-) show\n"
            "sizeR lineWidth 2 div n_seq 1 sub fontsize mul add moveto\n"
            "(-3\') show\n"
            "dirLen 0.5 mul neg lineWidth 2 div neg n_seq fontsize mul sub moveto\n"
            "(3\'-) show\n"
            "sizeS lineWidth 2 div neg n_seq fontsize mul sub moveto\n"
            "(-5\') show\n"
            "grestore\n\n"
            "%%draw names of sequences\n"
            "gsave\n"
            "/Helvetica-Bold findfont 8 scalefont setfont\n"
            "lenName dirLen add neg lineWidth n_seq 1 sub fontsize mul add 2 div moveto\n"
            "nameR show\n"
            "lenName dirLen add neg lineWidth neg n_seq 1 add fontsize mul sub 2 div moveto\n"
            "nameS show\n"
            "grestore\n\n"
            "%%draw arcs of secondary structure\n"
            "drawArcR\n"
            "drawArcS\n"
            "drawHybrid\n\n"
            "%%draw ruler\n"
            "ruler_r\n"
            "ruler_s\n\n"
            "showpage\n");

    fprintf(ptr_result,
            "%%%%Page: 1 1\n"
            "%%%%EOF\n");

	fclose(ptr_result);
}

void make_pair_table(const char *struc)
{
    char ch;
	int stck_sec[MAXLENGTH], stck_hyb[MAXLENGTH];
	int i, k;
    int top_sec = -1, top_hyb = -1, count = 0;

	for(i=0; i<MAXLENGTH; i++)
		stck_sec[i] = stck_hyb[i] = 0;

    for(i=0; i<strlen(struc); i++) {
		ch = struc[i];
        
		switch(ch) {
			case '(':
				stck_sec[++top_sec] = i;
				break;
			case ')':
                k = stck_sec[top_sec--];
                
                if (i < r[0]) {
                    r[k+1] = i+1;
                    r[i+1] = k+1;
                }
                else {
                    s[k-r[0]] = i-r[0];
                    s[i-r[0]] = k-r[0];
                }
                
				break;
			case '[':
				stck_hyb[++top_hyb] = i;
				break;
            case ']':
				k = stck_hyb[top_hyb--];
				hybrid[k+1] = s[0]-i+r[0]+1;
                count++;
                
				break;
			default:
				break;
		}
	}
    
	return;
}
