#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


#define   NUM   500000

typedef struct {
  int i;
  int j;
  double p;
} plist;

int k = 0;
int doi = 5;
int top_hyb[5], corr_matrix[5][5];
double cov[5][5];   //sample covariance
double corr[5][5];

void draw_heat_figure(FILE *fp, int len, int cut_point, plist *pl, const char *name_r,
                      const char *name_s);

int main(int argc, char *argv[])
{
    int i, j, tmp, len;
    double pro_i, pro_j;
    int count = 0, cut_point = -1;
    FILE *fp_sample, *fp_profile, *fp;
    char *file_name, *pro_file_name, line[1024], structure[1024];
    char *name_r = "U4", *name_s = "U6";
    plist profile[50000];
    
    if (argc != 3) {
        fprintf(stderr, "Please give me file name to read!\n");
        exit(EXIT_FAILURE);
    }

    fp = fopen("./result/heat.eps", "w");
    if (!fp) {
        fprintf(stderr, "Error to write file %s\n.", "./result/heat.eps");
        exit(EXIT_FAILURE);
    }
    file_name = argv[1];
    fp_sample = fopen(file_name, "r");
    if (!fp_sample) {
        fprintf(stderr, "Error to open file %s\n.", file_name);
        exit(EXIT_FAILURE);
    }
    pro_file_name = argv[2];
    fp_profile = fopen(pro_file_name, "r");
    if (!fp_profile) {
        fprintf(stderr, "Error to open file %s\n.", pro_file_name);
        exit(EXIT_FAILURE);
    }
    
    if (!fgets(line, 1024, fp_sample)) {
        fprintf(stderr, "No data in %s!\n", file_name);
        exit(EXIT_FAILURE);
    }
    while (strcmp(line, "Top frequency of hybrids:\n") != 0)
        fgets(line, 1024, fp_sample);
   
    fgets(line, 1024, fp_sample);
    while (strcmp(line, "Correlations:\n") != 0) {
        sscanf(line, "Top %d:\n", &i);
        fgets(line, 1024, fp_sample);
        sscanf(line, "%*s%d\n", &top_hyb[i-1]);
        count++;

        fgets(line, 1024, fp_sample);
    }
    if (count < 5)
        doi = count;
	
	printf("top_hyb[]:\n");
	for (i=0; i<doi; i++)
		printf("%8d ", top_hyb[i]);
	printf("\n\n");

    fgets(line, 1024, fp_sample);
    for (i=doi-1; i>=0; i--) {
        fscanf(fp_sample, "%d", &tmp);

        for (j=doi-1; j>i; j--)
            fscanf(fp_sample, "%d", &tmp);
        for (j=i; j>=0; j--) {
            fscanf(fp_sample, "%d", &tmp);
            corr_matrix[i][j] = tmp;
            corr_matrix[j][i] = tmp;
        }
    }
    printf("The sample correlation matrix:\n");
    printf("%c", ' ');
    for (i=0; i<doi; i++)
        printf("%8d", doi-i);
    printf("\n");
    for (i=doi-1; i>=0; i--) {
        printf("%d", i+1);
        for (j=doi-1; j>=0; j--)
            printf("%8d", corr_matrix[i][j]);
        printf("\n");
    }
    printf("\n");

    /* calculate covariance */
    for (i=0; i<doi; i++)
        for (j=0; j<doi; j++) {
            pro_i = corr_matrix[i][i]/((double) NUM);
            pro_j = corr_matrix[j][j]/((double) NUM);
            cov[i][j] = corr_matrix[i][j]/((double) NUM)-pro_i*pro_j;
        }
	printf("Covariance:\n");
    printf("%c", ' ');
    for (i=0; i<doi; i++)
        printf("%10d", doi-i);
    printf("\n");
    for (i=doi-1; i>=0; i--) {
        printf("%d", i+1);
        for (j=doi-1; j>=0; j--)
            printf("%10f", cov[i][j]);
        printf("\n");
    }
	printf("\n");
            
    fgets(line, 1024, fp_sample);  //read '\n'

    /* calculate @corr[5][5] */
    for (i=0; i<doi; i++)
        for (j=0; j<doi; j++)
            corr[i][j] = cov[i][j]/sqrt(cov[i][i]*cov[j][j]);

	printf("Correlation:\n");
    printf("%c", ' ');
    for (i=0; i<doi; i++)
        printf("%10d", doi-i);
    printf("\n");
	for (i=doi-1; i>=0; i--) {
        printf("%d", i+1);
		for (j=doi-1; j>=0; j--)
			printf("%10f", corr[i][j]);
		printf("\n");
	}
    printf("\n");

    fgets(line, 1024, fp_sample);
    fgets(line, 1024, fp_sample);
    fscanf(fp_sample, "%s", structure);
    len = strlen(structure);
    for (i=0; i<strlen(structure); i++)
        if (structure[i] == '|') {
            cut_point = i+1;
            break;
        }

    /* read profile data from profile ps file */
    if (!fgets(line, 1024, fp_profile)) {
        fprintf(stderr, "No data in %s!\n", pro_file_name);
        exit(EXIT_FAILURE);
    }
    while (strcmp(line, "%data starts here\n") != 0)
        fgets(line, 1024, fp_profile);
    
    fgets(line, 1024, fp_profile);
    while (strcmp(line, "showpage\n") != 0) {
        sscanf(line, "%d %d %lf %*s\n", &profile[k].i, &profile[k].j, &profile[k].p);
        k++;
        fgets(line, 1024, fp_profile);
    }
    
    draw_heat_figure(fp, len, cut_point, profile, name_r, name_s);

    fprintf(stdout, "The heat figure is generated successfully in \"./result/heat.eps\"!\n");
    fclose(fp_sample);
    fclose(fp_profile);
    fclose(fp);
    
    return EXIT_SUCCESS;
}

void draw_heat_figure(FILE *fp, int len, int cut_point, plist *pl, const char *name_r,
                      const char *name_s)
{
    int i, j;
    
    fprintf(fp,
            "%%!PS-Adobe-3.0 EPSF-3.0\n"
            "%%%%BoundingBox: 0 0 200 300\n\n\n"
            "%%%%BeginProlog\n"
            "/DPdict 100 dict def\n\n"
            "DPdict begin\n"
            "/logscale false def\n"
            "/lpmin 1e-05 log def\n\n"
            "/box { %%size x y box - draws box centered on x,y\n"
            "   3 2 roll 1 exch sub setgray 1 3 1 roll\n"
            "   2 index 0.5 mul sub            %% x -= 0.5\n"
            "   exch 2 index 0.5 mul sub exch  %% y -= 0.5\n"
            "   3 -1 roll dup rectfill\n"
            "} bind def\n\n"
            "/ubox {\n"
            "   logscale {\n"
            "      log dup add lpmin div 1 exch sub dup 0 lt { pop 0 } if\n"
            "   } if\n"
            "   3 1 roll\n"
            "   exch len exch sub 1 add box\n"
            "} bind def\n"
            "\n"
            "/lbox {\n"
            "   3 1 roll\n"
            "   len exch sub 1 add box\n"
            "} bind def\n"
            "\n"
            "end\n"
            "%%%%EndProlog\n\n\n"
            "DPdict begin\n"
            "\n"
            "/fwidth 200 def\n"
            "/fheight 300 def\n\n");

    fprintf(fp, "/name_r (%s) def\n", name_r);
    fprintf(fp, "/name_s (%s) def\n", name_s);
    fprintf(fp, "/len %d def\n", len);
    if (cut_point > 0)
        fprintf(fp, "/cutpoint %d def\n", cut_point);
    fprintf(fp, "/disp 2 def\n\n");

    fprintf(fp,
            "-90 rotate\n"
            "fheight neg 0 translate\n"
            "\n"
            "gsave\n"
            "30 30 translate\n"
            "/Helvetica findfont 8 scalefont setfont\n"
            "0.08 setlinewidth\n"
            "\n"
            "%%draw box\n"
            "gsave\n"
            "/Helvetica findfont 0.95 scalefont setfont\n"
            "\n"
            "newpath\n"
            "0 0 moveto\n"
            "0 cutpoint disp add lineto\n"
            "len cutpoint sub disp add cutpoint disp add lineto\n"
            "len cutpoint sub disp add 0 lineto\n"
            "closepath\n"
            "\n"
            "stroke\n"
            "grestore\n"
            "\n"
            "%%draw names and direction\n"
            "gsave\n"
            "-12 cutpoint disp add 2 div name_r stringwidth pop 2 div add moveto gsave -90 rotate name_r show grestore\n"
            "-12 cutpoint disp add moveto gsave -90 rotate (5\') show grestore\n"
            "-12 2 moveto gsave -90 rotate (3\') show grestore\n"
            "len cutpoint sub disp add 2 div name_s stringwidth pop 2 div sub -12 moveto name_s show\n"
            "-2 -12 moveto (5\') show\n"
            "len cutpoint sub disp add 2 sub -12 moveto (3\') show\n"
            "grestore\n"
            "\n"
            "%%draw ruler\n"
            "gsave\n"
            "/Helvetica findfont 0.95 scalefont setfont\n"
            "\n"
            "/step 0 def\n"
            "{\n"
            "	-3 step moveto 3 0 rlineto\n"
            "	/step step 10 add def\n"
            "	step cutpoint gt { exit } if\n"
            "} loop\n"
            "-3 cutpoint disp add moveto 3 0 rlineto\n"
            "\n"
            "/step 0 def\n"
            "{\n"
            "	step -3 moveto 0 3 rlineto\n"
            "	/step step 10 add def\n"
            "	step len cutpoint sub gt { exit } if\n"
            "} loop\n"
            "len cutpoint sub disp add -3 moveto 0 3 rlineto\n"
            "\n"
            "stroke\n"
            "grestore\n"
            "\n"
            "cutpoint neg 1 add len cutpoint sub neg 1 sub translate\n\n");

    fprintf(fp, "%%start data\n");
	for (i=0; i<k; i++)
		if (pl[i].i != pl[i].j)
			fprintf(fp, "%d %d %1.9f ubox\n", pl[i].i, pl[i].j, pl[i].p);

    fprintf(fp,
            "\n0 0 0 setrgbcolor\n"
            "grestore\n"
            "%%---------- end data ----------------\n\n");

    /* -------------------- draw heat figure ------------------- */
    fprintf(fp,
            "%%----------------- draw heat figure ---------------------\n"
            "gsave\n"
            "/boxwidth 0.9 def\n"
            "/boxheight 0.9 def \n"
            "\n"
            "%% Data\n"
            "/XLABELS [\n");
    for (i=0; i<5; i++)
        fprintf(fp, "\t[ %d (%d) ]\n", i, i+1);
    fprintf(fp, "] def\n\n"
            "/YLABELS [\n");
    for (i=0; i<5; i++)
        fprintf(fp, "\t[ %d (%d) ]\n", i, i+1);
    fprintf(fp, "] def\n\n"
            "/THINGS [\n");
    for (i=0; i<doi; i++)
        fprintf(fp, "\t[ %d %d  %9f ]\n", i, i, corr[i][i]);
    for (i=0; i<doi-1; i++)
        for (j=i+1; j<doi; j++) {
            fprintf(fp, "\t[ %d %d  %9f ]\n", i, j, corr[i][j]);
            fprintf(fp, "\t[ %d %d  %9f ]\n", j, i, corr[j][i]);
        }
    fprintf(fp, "] def\n\n");

    fprintf(fp,
            "%% PROGRAM\n"
            "/max { 2 copy lt { exch } if pop } bind def\n"
            "/width   { XLABELS length } bind def\n"
            "/height  { YLABELS length } bind def \n"
            "\n"
            "width height max /size exch def\n"
            "\n"
            "/rectangle {%% x y width height rectangle\n"
            "            %% draws a box centered at x,y with width and height as given.\n"
            "	    /hh exch def\n"
            "	    /ww exch def\n"
            "	    newpath\n"
            "        size exch sub \n"
            "	    moveto\n"
            "	    ww 0.5 mul neg hh 0.5 mul neg rmoveto \n"
            "	    ww 0 rlineto 0 hh rlineto ww neg 0 rlineto\n"
            "	    closepath fill\n"
            "} bind def\n\n"
            "/label {\n"
            "	/hh exch def\n"
            "	/ww exch def\n"
            "    size exch sub \n"
            "	moveto\n"
            "	(-) stringwidth pop 2 div neg -0.25 rmoveto\n"
            "	(-) show\n"
            "} bind def\n\n"
            "/heatcolor {\n"
            "    dup 0 gt {\n"
            "		1 exch sub dup 1 3 1 roll setrgbcolor\n"
            "	} {\n"
            "		neg 1 exch sub dup 1 setrgbcolor\n"
            "	} ifelse\n"
            "} bind def \n"
            "\n"
            "/BX {\n"
            "    heatcolor\n"
            "    boxwidth boxheight rectangle\n"
            "} bind def\n"
            "\n"
            "/HLINE {\n"
            "     height exch sub 0.5 add dup -0.6 exch  moveto width 0.4 sub exch\n"
            "     lineto stroke\n"
            "} def \n"
            "\n"
            "30 len cutpoint sub add 40 add 50 translate \n"
            "13 dup scale\n"
            "/Helvetica findfont 0.9 scalefont setfont\n"
            "0.005 size mul setlinewidth\n"
            "newpath\n"
            " -0.6  0.4 moveto\n"
            " -0.6 height 0.6 add lineto\n"
            "width 0.4 sub height 0.6 add lineto\n"
            "width 0.4 sub 0.4  lineto closepath\n"
            "stroke\n"
            "\n"
            "0 0 0 setrgbcolor\n"
            "\n"
            "YLABELS {\n"
            "   aload pop\n"
            "   exch height exch sub -1 exch moveto\n"
            "   dup stringwidth pop neg -0.25 rmoveto\n"
            "   show\n"
            "} forall \n"
            "\n"
            "\n"
            "XLABELS {\n"
            "   aload pop \n"
            "   exch height 1 add moveto\n"
            "   dup stringwidth pop 0.2 sub neg -0.25 rmoveto\n"
            "   show\n"
            "} forall\n"
            "\n"
            "THINGS { aload pop BX } forall\n"
            "\n"
            "stroke\n"
            "0 0 0 setrgbcolor\n");
    fprintf(fp, "width 1.0 sub 2 div (%s-%s) stringwidth pop 2 div sub -0.5 moveto\n", name_r, name_s);
    fprintf(fp, "(%s-%s) show\n", name_r, name_s);
    for (i=0; i<doi; i++)
        for (j=0; j<doi; j++)
            if (corr[i][j] < 0.0)
                fprintf(fp, "%d %d boxwidth boxheight label\n", i, j);
    fprintf(fp,
            "\n"
            "0 0 0 setrgbcolor\n"
            "stroke\n"
            "grestore\n"
            "\n"
            "gsave\n"
            "/boxwidth 0.9 def\n"
            "/boxheight 0.9 def \n"
            "\n"
            "%%Data\n"
            "/XLABELS [\n"
            "    [ 0 ( ) ]\n"
            "] def\n\n\n"
            "/YLABELS [\n");

    for (i=0; i<=20; i++)
        fprintf(fp, "\t[ %d (%.1f) ]\n", i, 1.0-0.1*i);
    fprintf(fp,
            "] def\n\n"
            "/THINGS [\n");
    for (i=0; i<=20; i++)
        fprintf(fp, "\t[ 0 %d  %.1f ]\n", i, 1.0-0.1*i);
    fprintf(fp, "] def\n\n");

    fprintf(fp,
            "width height max /size exch def\n"
            "\n"
            "270 30 translate\n"
            "6 dup scale\n"
            "/Courier findfont 0.8 scalefont setfont\n"
            "0.001 size mul setlinewidth\n"
            "newpath\n"
            " -0.6  0.4 moveto\n"
            " -0.6 height 0.6 add lineto\n"
            "width 0.4 sub height 0.6 add lineto\n"
            "width 0.4 sub 0.4 lineto\n"
            "closepath\n"
            "stroke\n"
            "\n"
            "0 0 0 setrgbcolor\n"
            "\n"
            "YLABELS {\n"
            "   aload pop\n"
            "   exch height exch sub -1 exch moveto\n"
            "   dup stringwidth pop neg -0.25 rmoveto\n"
            "   show\n"
            "} forall \n"
            "\n"
            "\n"
            "XLABELS {\n"
            "   aload pop \n"
            "   exch height 1 add moveto\n"
            "   dup stringwidth pop neg -0.25 rmoveto\n"
            "   show\n"
            "} forall\n"
            "\n"
            "THINGS { aload pop BX } forall\n"
            "\n"
            "stroke\n"
            "0 0 0 setrgbcolor\n"
            "grestore\n\n");
            
    fprintf(fp, "showpage\n"
            "end\n"
            "%%%%EOF");
    
    return;
}
