#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "energy.h"
#include "pfunc.h"


paramT *P;

int num_align, len_r, len_s;
int **aSR, **aSS;
int *indxR, *indxS;
double *pair_scoreR, *pair_scoreS, **pair_scoreHy;
int **pair_typeR, **pair_typeS, ***pair_typeHy;

double *QbR, *QbsR, *QR, *QmR, *QbS, *QbsS, *QS, *QmS;
double *QfR, *QfS;
/**
 *  0: eea   1: eeb   2: me    3: em    4: fe   5: ef  6: mm  7: mf  8: fm  9: ff
 * 10: eka  11: ekb  12: mk   13: fk   14: kea  15: keb
 * 16: km   17: kf   18: kka  19: kkb
 */
double **QRT[20];
double **QDT[20];
/**
 * tight structure
 * 0: ee  1: em/me  2: ef/fe  3: ek/ke
 */
double **QTu[4], **QTsu[4];
double **QTd[4], **QTsd[4];
double **QTe, **QTse;
double **Bond[4];
double Qany;

/**
 * structure constraint: BP[i]
 * -1---'-':  base which must be paired with some base in the same strand
 * -2---'^':  base which must be paired with some base in the other strand
 * -3---'<':  base which must be paired with j<i in the same strand
 * -4---'>':  base which must be paired with j>i in the same strand
 * -5---'x':  base which pairs on base
 * positive numbers:  base i pairs base BP[i]
 */
int *BP;
double cv_fact = 1.0;
double nc_fact = 1.0;


void reverse_align(char **seqs, char **inv_seqs)
{
	int i, j, len;

	len = strlen(seqs[1]);
    for (i=1; i<=num_align; i++) {
        for (j=len-1; j>=0; j--)
            inv_seqs[i][len-j-1] = seqs[i][j];
        inv_seqs[i][len] = '\0';
    }

    return;
}

void erase_space(char *seq)
{
    int i, k;
    char seq_tmp[1024];

    i = k = 0;
    while(seq[i] != '\0') {
        if (seq[i] == ' ') {
            i++;
            continue;
        }
        seq_tmp[k++] = seq[i++];
    }
    seq_tmp[k] = '\0';
    strcpy(seq, seq_tmp);
}

void shrink_align(char ***ptr_seqs_r, char ***ptr_seqs_s)
{
    char **seqs_r, **seqs_s;
    int i, j, t, k;
    int n_ph, pos_ph[500];
    int length_r, length_s;

    seqs_r = *ptr_seqs_r;
    seqs_s = *ptr_seqs_s;
    length_r = len_r;
    length_s = len_s;

    /* print the sequence before */
    printf("shrink before:\n");
    for (k=1; k<=num_align; k++) {
        for (i=0; i<len_r; i++)
            printf("%c", seqs_r[k][i]);
        printf("\n");
    }
    for (k=1; k<=num_align; k++) {
        for (i=0; i<len_s; i++)
            printf("%c", seqs_s[k][i]);
        printf("\n");
    }
    printf("\n");
    
    //R
    t = 0;
    memset(pos_ph, 0, 500);
    for (i=0; i<len_r; i++) {
        n_ph = 0;
        
        for (k=1; k<=num_align; k++)
            if (seqs_r[k][i] == '.')
                n_ph++;

        if (n_ph == num_align)
            pos_ph[t++] = i;
    }
    for (i=t-1; i>=0; i--) {
        for (k=1; k<=num_align; k++) {
            for (j=pos_ph[i]+1; j<len_r; j++)
                seqs_r[k][j-1] = seqs_r[k][j];
            
            seqs_r[k][len_r-1] = '\0';
        }

        len_r--;
    }

    //S
    t = 0;
    memset(pos_ph, 0, 500);
    for (i=0; i<len_s; i++) {
        n_ph = 0;
        
        for (k=1; k<=num_align; k++)
            if (seqs_s[k][i] == '.')
                n_ph++;

        if (n_ph == num_align)
            pos_ph[t++] = i;
    }
    for (i=t-1; i>=0; i--) {
        for (k=1; k<=num_align; k++) {
            for (j=pos_ph[i]+1; j<len_s; j++)
                seqs_s[k][j-1] = seqs_s[k][j];

            seqs_s[k][len_s-1] = '\0';
        }

        len_s--;
    }

    if (len_r < length_r)
        printf("Shrink length of alignmet R from %d to %d.\n", length_r, len_r);
    if (len_s < length_s)
        printf("Shrink length of alignmet R from %d to %d.\n", length_s, len_s);
    printf("\n");

    /* print the sequences after */
    printf("shrink after:\n");
    for (k=1; k<=num_align; k++) {
        for (i=0; i<len_r; i++)
            printf("%c", seqs_r[k][i]);
        printf("\n");
    }
    for (k=1; k<=num_align; k++) {
        for (i=0; i<len_s; i++)
            printf("%c", seqs_s[k][i]);
        printf("\n");
    }
    
    return;    
}

void read_align(char ***ptr_seqs_r, char ***ptr_seqs_s, const char *file_name_r, const char *file_name_s)
{
    int i, cursor = 0;
    FILE *fp_input_r, *fp_input_s;
    char *ptr, line[MAXLENGTH], seq[MAXLENGTH];

    fp_input_r = fopen(file_name_r, "r");
    if (!fp_input_r) {
        fprintf(stderr, "Can not open file %s.\n", file_name_r);
        exit(EXIT_FAILURE);
    }
    fp_input_s = fopen(file_name_s, "r");
    if (!fp_input_s) {
        fprintf(stderr, "Can not open file %s.\n", file_name_s);
        exit(EXIT_FAILURE);
    }
    
    memset(seq, '\0', MAXLENGTH);
    memset(line, '\0', MAXLENGTH);

    //R
    if (!fgets(line, MAXLENGTH, fp_input_r)) {   //read the description line starting with >
        fprintf(stderr, "No data in alignment R\n");
        exit(EXIT_FAILURE);
    }
    if (line[0] != '>') {
        fprintf(stderr, "Not FASTA format!\n");
        exit(EXIT_FAILURE);
    }

    if (fgets(line, MAXLENGTH, fp_input_r) == NULL) {
        fprintf(stderr, "No sequence in R found!\n");
        exit(EXIT_FAILURE);
    }
    while (line[0] != '>') {
        erase_space(line);
        strcat(seq, line);
        if (!fgets(line, MAXLENGTH, fp_input_r))     //only one sequence
            break;
    }
    len_r = strlen(seq)-1;
    memset(seq, '\0', MAXLENGTH);
    cursor++;
    
    while (fgets(line, MAXLENGTH, fp_input_r) != NULL) {
        if (line[0] != '>') {   // a new sequence
            erase_space(line);
            strcat(seq, line);
            continue;
        }
        
        if (len_r != strlen(seq)-1) {
            fprintf(stderr, "The numbers of columns of Alignment R are not equal.\n");
            exit(EXIT_FAILURE);
        }
        memset(seq, '\0', MAXLENGTH);
        cursor++;
    }
    if (cursor>1 && len_r!=strlen(seq)-1) {
        fprintf(stderr, "The numbers of columns of Alignment R are not equal.\n");
        exit(EXIT_FAILURE);
    }
    if (cursor > 1)
        cursor++;

    //S
    memset(seq, '\0', MAXLENGTH);
    if (!fgets(line, MAXLENGTH, fp_input_s)) {   //read the description line starting with >
        fprintf(stderr, "No data in alignment S\n");
        exit(EXIT_FAILURE);
    }
    if (line[0] != '>') {
        fprintf(stderr, "Not FASTA format!\n");
        exit(EXIT_FAILURE);
    }

    if (fgets(line, MAXLENGTH, fp_input_s) == NULL) {
        fprintf(stderr, "No sequence in S found!\n");
        exit(EXIT_FAILURE);
    }
    while (line[0] != '>') {
        erase_space(line);
        strcat(seq, line);
        if (!fgets(line, MAXLENGTH, fp_input_s))
            break;
    }
    len_s = strlen(seq)-1;
    memset(seq, '\0', MAXLENGTH);
    num_align++;
    
    while (fgets(line, MAXLENGTH, fp_input_s) != NULL) {
        if (line[0] != '>') {   // a new sequence
            erase_space(line);
            strcat(seq, line);
            continue;
        }
        
        if (len_s != strlen(seq)-1) {
            fprintf(stderr, "The numbers of columns of Alignment S are not equal.\n");
            exit(EXIT_FAILURE);
        }
        memset(seq, '\0', MAXLENGTH);
        num_align++;
    }
    if (num_align>1 && len_s!=strlen(seq)-1) {
        fprintf(stderr, "The numbers of columns of Alignment S are not equal.\n");
        exit(EXIT_FAILURE);
    }
    if (num_align > 1)
        num_align++;

    if (cursor != num_align) {
        fprintf(stderr, "The number of the sequences of the two alignments is not equal.\n");
        exit(EXIT_FAILURE);
    }

    /* ----------------- allocate memory for @seqs_r and @seqs_s --------------------- */
    //R
    *ptr_seqs_r = (char **) space(sizeof(char *) * (num_align+1));
    ptr = (char *) space(sizeof(char) * (num_align+1) * (len_r+2));
    for (i=0; i<num_align+1; i++) {
        (*ptr_seqs_r)[i] = ptr;
        ptr += len_r+2;
    }
    //S
    *ptr_seqs_s = (char **) space(sizeof(char *) * (num_align+1));
    ptr = (char *) space(sizeof(char) * (num_align+1) * (len_s+2));
    for (i=0; i<num_align+1; i++) {
        (*ptr_seqs_s)[i] = ptr;
        ptr += len_s+2;
    }

    fseek(fp_input_r, 0L, SEEK_SET);
    fseek(fp_input_s, 0L, SEEK_SET);
    /* ---------------------- read data from the two input files --------------------- */
    //R
    i = 0;
    memset(seq, '\0', MAXLENGTH);
    while (fgets(line, MAXLENGTH, fp_input_r) != NULL) {
        if (line[0] == '>') {
            strcpy((*ptr_seqs_r)[i++], seq);
            (*ptr_seqs_r)[i-1][len_r] = '\0';
            memset(seq, '\0', MAXLENGTH);
            fgets(line, MAXLENGTH, fp_input_r);
        }

        line[strlen(line)-1] = '\0';
        erase_space(line);
        strcat(seq, line);
    }
    strcpy((*ptr_seqs_r)[i++], seq);
    (*ptr_seqs_r)[i-1][len_r] = '\0';
    memset(seq, '\0', MAXLENGTH);
    //S
    i = 0;
    memset(seq, '\0', MAXLENGTH);
    while (fgets(line, MAXLENGTH, fp_input_s) != NULL) {
        if (line[0] == '>') {
            strcpy((*ptr_seqs_s)[i++], seq);
            (*ptr_seqs_s)[i-1][len_s] = '\0';
            memset(seq, '\0', MAXLENGTH);
            fgets(line, MAXLENGTH, fp_input_s);
        }

        line[strlen(line)-1] = '\0';
        erase_space(line);
        strcat(seq, line);
    }
    strcpy((*ptr_seqs_s)[i++], seq);
    (*ptr_seqs_s)[i-1][len_s] = '\0';

    if (do_shrink)
        shrink_align(ptr_seqs_r, ptr_seqs_s);

    fclose(fp_input_r);
    fclose(fp_input_s);
}

char *read_align_constraint(const char *file_name)
{
    FILE *fp;
    char ch, line[2048];
    char *cons_string;  //constraint string
    int i;

    fp = fopen(file_name, "r");
    if (!fp) {
        fprintf(stderr, "Can not open file %s.\n", file_name);
        exit(EXIT_FAILURE);
    }
    
    cons_string = (char *) space(sizeof(char)*(len_r+len_s+2));

    /*
    printf("Please input constraints using the following notation:\n");
    printf(". : no constraint at all\n");
    printf("- : paired with some base on the same strand\n");
    printf("^ : paired with some base on the other strand\n");
    printf("x : base must not pair\n");
    printf("< : base i is paired with a base j<i on the same strand\n");
    printf("> : base i is paired with a base j>i on the same strand\n");
    printf("matching brackets ( ): base i pairs base j, where i and j are on the same strand\n");
    printf("matching brackets [ ]: base i pairs base j, where i and j are on different strand\n");
    printf("using | to separate constraints into two parts\n");

    while(flag) {
        int cntn = 0;   //check if continue inputing constraint string or not
        
        printf("#");
        fgets(line, 1024, stdin);
        
        //check its validity
        if ((strlen(line)-1) != (len_r+len_s+1)) {
            fprintf(stderr, "wrong length of constraints. The length should be %d. Try again.\n",
                    len_r+len_s+1);
            continue;
        }
        if (line[len_r] != '|') {
            fprintf(stderr, "The separator | not found in position %d. Try again\n", len_r+1);
            continue;
        }
        for (i=0; i<len_r; i++) {
            ch=line[i];
                
            if (ch!='.' && ch!='-' && ch!='^' && ch!='x' && ch!='<'
                && ch!='>' && ch!='(' && ch!=')' && ch!='[') {
                fprintf(stderr, "invalid characters found in R. Try again.\n");
                cntn = 1;
                break;
            }
        }
        if (cntn == 1)
            continue;
            
        for (i=len_r+1; i<len_r+len_s+1; i++) {
            ch=line[i];
                
            if (ch!='.' && ch!='-' && ch!='^' && ch!='x' && ch!='<'
                && ch!='>' && ch!='(' && ch!=')' && ch!=']') {
                fprintf(stderr, "invalid characters found in S. Try again.\n");
                cntn = 1;
                break;
            }
        }
        if (cntn == 1)
            continue;
            
        flag=0;
    }   //end while (flag)
    */

    memset(line, '\0', 2048);
    fgets(line, 2048, fp);
    if (strlen(line) == 0) {
        fprintf(stderr, "No constraint string found!\n");
        exit(EXIT_FAILURE);
    }
    if ((strlen(line)-1) != (len_r+len_s+1)) {
        fprintf(stderr, "wrong length of constraints. The length should be %d. Please check the constraint.\n",
                len_r+len_s+1);
        exit(EXIT_FAILURE);
    }
    if (line[len_r] != '|') {
        fprintf(stderr, "The separator | not found in position %d.\n", len_r+1);
        exit(EXIT_FAILURE);
    }
    for (i=0; i<len_r+len_s+1; i++) {
        ch=line[i];
        
        if (ch!='.' && ch!='-' && ch!='^' && ch!='x' && ch!='<'
            && ch!='>' && ch!='(' && ch!=')' && ch!='[' && ch!='|') {
            fprintf(stderr, "invalid characters found in constraint string.\n");
            exit(EXIT_FAILURE);
        }
    }
    
    strncpy(cons_string, line, len_r+len_s+1);
    cons_string[len_r+len_s+1] = '\0';

    fclose(fp);
    
    return cons_string;
}

void initialize_params()
{
    P = scale_parameters();
}

void initialize_matrix()
{
	int i, j, *ptr;
	double *temp;

    aSR = (int **) space(sizeof(int *) * (num_align+1));
    ptr = (int *) space(sizeof(int) * (num_align+1) * (len_r+1));
    for (i=0; i<(num_align+1); i++) {
        aSR[i] = ptr;
        ptr += len_r+1;
    }
    
    aSS = (int **) space(sizeof(int *) * (num_align+1));
    ptr = (int *) space(sizeof(int) * (num_align+1) * (len_s+1)) ;
    for (i=0; i<(num_align+1); i++) {
        aSS[i] = ptr;
        ptr += len_s+1;
    }
    
	indxR = (int *) space(sizeof(int) * (len_r+1));
	indxS = (int *) space(sizeof(int) * (len_s+1));

    pair_scoreR = (double *) space(sizeof(double)*(len_r*(len_r+1)/2+2));
    pair_scoreS = (double *) space(sizeof(double)*(len_s*(len_s+1)/2+2));
    pair_scoreHy = (double **) space(sizeof(double *)*(len_r+1));
    temp = (double *) space(sizeof(double)*(len_r+1)*(len_s+1));
    for (i=0; i<(len_r+1); i++) {
        pair_scoreHy[i] = temp;
        temp += len_s+1;
    }
    
    pair_typeR = (int **) space(sizeof(int *) * (num_align+2));
    ptr = (int *) space(sizeof(int) * (num_align+2) * (len_r*(len_r+1)/2+2));
    for (i=0; i<num_align+2; i++) {
        pair_typeR[i] = ptr;
        ptr += len_r*(len_r+1)/2+2;
    }

    pair_typeS = (int **) space(sizeof(int *) * (num_align+2));
    ptr = (int *) space(sizeof(int) * (num_align+2) * (len_s*(len_s+1)/2+2));
    for (i=0; i<num_align+2; i++) {
        pair_typeS[i] = ptr;
        ptr += len_s*(len_s+1)/2+2;
    }

    pair_typeHy = (int ***) space(sizeof(int **) * (num_align+2));
    for (i=0; i<num_align+2; i++) {
        pair_typeHy[i] = (int **) space(sizeof(int *) * (len_r+1));
        
        ptr = (int *) space(sizeof(int) * (len_r+1) * (len_s+1));
        for (j=0; j<len_r+1; j++) {
            pair_typeHy[i][j] = ptr;
            ptr += len_s+1;
        }
    }
    
	QbR = (double *) space(sizeof(double) * (len_r*(len_r+1)/2+2));
    QbsR = (double *) space(sizeof(double) * (len_r*(len_r+1)/2+2));
	QR = (double *) space(sizeof(double) * (len_r*(len_r+1)/2+2));
	QmR = (double *) space(sizeof(double) * (len_r*(len_r+1)/2+2));
	QfR = (double *) space(sizeof(double) * (len_r*(len_r+1)/2+2));
	QbS = (double *) space(sizeof(double) * (len_s*(len_s+1)/2+2));
    QbsS = (double *) space(sizeof(double) * (len_s*(len_s+1)/2+2));
	QS = (double *) space(sizeof(double) * (len_s*(len_s+1)/2+2));
	QmS = (double *) space(sizeof(double) * (len_s*(len_s+1)/2+2));
	QfS = (double *) space(sizeof(double) * (len_s*(len_s+1)/2+2));
    
	for (i=1; i<=len_r; i++)
		indxR[i] = (i*(i-1)) >> 1;        /* n(n-1)/2 */
	for (i=1; i<=len_s; i++)
		indxS[i] = (i*(i-1)) >> 1;        /* n(n-1)/2 */

	for (j=0; j<20; j++) {
        QDT[j] = (double **) space(sizeof(double *) * (len_r*(len_r+1)/2+2));
        temp = (double *) space(sizeof(double) * (len_r*(len_r+1)/2+2) * (len_s*(len_s+1)/2+2));
        for (i=0; i<len_r*(len_r+1)/2+2; i++){
            QDT[j][i] = temp;
            temp += len_s*(len_s+1)/2+2;
        }   
	}
    
	for (j=0; j<20; j++) {
        QRT[j] = (double **) space(sizeof(double *) * (len_r*(len_r+1)/2+2));
        temp = (double *) space(sizeof(double) * (len_r*(len_r+1)/2+2) * (len_s*(len_s+1)/2+2));
        for (i=0; i<len_r*(len_r+1)/2+2; i++){
            QRT[j][i] = temp;
            temp += len_s*(len_s+1)/2+2;
        } 
	}

	for (j=0; j<4; j++) {
        QTu[j] = (double **) space(sizeof(double  * ) * (len_r*(len_r+1)/2+2));
        temp = (double *) space(sizeof(double) * (len_r*(len_r+1)/2+2) * (len_s*(len_s+1)/2+2));
        for (i=0; i<len_r*(len_r+1)/2+2; i++){
            QTu[j][i] = temp;
            temp += len_s*(len_s+1)/2+2;
        }

        QTsu[j] = (double **) space(sizeof(double *) * (len_r*(len_r+1)/2+2));
        temp = (double *) space(sizeof(double) * (len_r*(len_r+1)/2+2) * (len_s*(len_s+1)/2+2));
        for (i=0; i<len_r*(len_r+1)/2+2; i++){
            QTsu[j][i] = temp;
            temp += len_s*(len_s+1)/2+2;
        }

        QTd[j] = (double **) space(sizeof(double *) * (len_r*(len_r+1)/2+2));
        temp = (double *) space(sizeof(double) * (len_r*(len_r+1)/2+2) * (len_s*(len_s+1)/2+2));
        for (i=0; i<len_r*(len_r+1)/2+2; i++){
            QTd[j][i] = temp;
            temp += len_s*(len_s+1)/2+2;
        }

        QTsd[j] = (double **) space(sizeof(double *) * (len_r*(len_r+1)/2+2));
        temp = (double *) space(sizeof(double) * (len_r*(len_r+1)/2+2) * (len_s*(len_s+1)/2+2));
        for (i=0; i<len_r*(len_r+1)/2+2; i++){
            QTsd[j][i] = temp;
            temp += len_s*(len_s+1)/2+2;
        }
	}
    
	QTe = (double **) space(sizeof(double *) * (len_r*(len_r+1)/2+2));
	temp = (double *) space(sizeof(double) * (len_r*(len_r+1)/2+2) * (len_s*(len_s+1)/2+2));
    for (i=0; i<len_r*(len_r+1)/2+2; i++){
	    QTe[i] = temp;
	    temp += len_s*(len_s+1)/2+2;
    }

    QTse = (double **) space(sizeof(double *) * (len_r*(len_r+1)/2+2));
    temp = (double *) space(sizeof(double) * (len_r*(len_r+1)/2+2) * (len_s*(len_s+1)/2+2));
    for (i=0; i<len_r*(len_r+1)/2+2; i++){
        QTse[i] = temp;
	    temp += len_s*(len_s+1)/2+2;
    }
    
	for (j=0; j<4; j++) {
        Bond[j] = (double **) space(sizeof(double *) * (len_r*(len_r+1)/2+2));
        temp = (double *) space(sizeof(double) * (len_r*(len_r+1)/2+2) * (len_s*(len_s+1)/2+2));
        for (i=0; i<len_r*(len_r+1)/2+2; i++){
            Bond[j][i] = temp;
            temp += len_s*(len_s+1)/2+2;
        }
	}
}

void clean_matrix(void)
{
    int j;
    
    memset(aSR[0], 0, (num_align+1)*(len_r+1)*sizeof(int));
    memset(aSS[0], 0, (num_align+1)*(len_s+1)*sizeof(int));
	
    memset(indxR, 0, (len_r+1)*sizeof(int));
    memset(indxS, 0, (len_s+1)*sizeof(int));
    
    memset(pair_scoreR, 0, (len_r*(len_r+1)/2+2)*sizeof(double));
    memset(pair_scoreS, 0, (len_s*(len_s+1)/2+2)*sizeof(double));
    memset(pair_scoreHy[0], 0, (len_r+1)*(len_s+1)*sizeof(double));
    
    memset(pair_typeR[0], 0, (num_align+2)*(len_r*(len_r+1)/2+2)*sizeof(int));
    memset(pair_typeS[0], 0, (num_align+2)*(len_s*(len_s+1)/2+2)*sizeof(int));
    for (j=0; j<(num_align+2); j++)
        memset(pair_typeHy[j][0], 0, (len_r+1)*(len_s+1)*sizeof(int));
	
    memset(QbR, 0, sizeof(double)*(len_r*(len_r+1)/2+2));
    memset(QbsR, 0, sizeof(double)*(len_r*(len_r+1)/2+2));
    memset(QR, 0, sizeof(double)*(len_r*(len_r+1)/2+2));
    memset(QmR, 0, sizeof(double)*(len_r*(len_r+1)/2+2));
    memset(QfR, 0, sizeof(double)*(len_r*(len_r+1)/2+2));
    
    memset(QbS, 0, sizeof(double)*(len_s*(len_s+1)/2+2));
    memset(QbsS, 0, sizeof(double)*(len_s*(len_s+1)/2+2));
    memset(QS, 0, sizeof(double)*(len_s*(len_s+1)/2+2));
    memset(QmS, 0, sizeof(double)*(len_s*(len_s+1)/2+2));
    memset(QfS, 0, sizeof(double)*(len_s*(len_s+1)/2+2));
    
	for (j=1; j<=len_r; j++)
		indxR[j] = (j*(j-1)) >> 1;        /* n(n-1)/2 */
	for (j=1; j<=len_s; j++)
		indxS[j] = (j*(j-1)) >> 1;        /* n(n-1)/2 */
	
    for (j=0; j<20; j++)
        memset(QDT[j][0], 0, sizeof(double)*(len_r*(len_r+1)/2+2)*(len_s*(len_s+1)/2+2));
    for (j=0; j<20; j++)
        memset(QRT[j][0], 0, sizeof(double)*(len_r*(len_r+1)/2+2)*(len_s*(len_s+1)/2+2));
    for (j=0; j<4; j++) {
        memset(QTu[j][0], 0, sizeof(double)*(len_r*(len_r+1)/2+2)*(len_s*(len_s+1)/2+2));
        memset(QTsu[j][0], 0, sizeof(double)*(len_r*(len_r+1)/2+2)*(len_s*(len_s+1)/2+2));
        memset(QTd[j][0], 0, sizeof(double)*(len_r*(len_r+1)/2+2)*(len_s*(len_s+1)/2+2));
        memset(QTsd[j][0], 0, sizeof(double)*(len_r*(len_r+1)/2+2)*(len_s*(len_s+1)/2+2));
    }
    memset(QTe[0], 0, sizeof(double)*(len_r*(len_r+1)/2+2)*(len_s*(len_s+1)/2+2));
    memset(QTse[0], 0, sizeof(double)*(len_r*(len_r+1)/2+2)*(len_s*(len_s+1)/2+2));
    for (j=0; j<4; j++)
        memset(Bond[j][0], 0, sizeof(double)*(len_r*(len_r+1)/2+2)*(len_s*(len_s+1)/2+2));

    return;
}

void free_matrix()
{
	int i,j;

    free(aSR[0]); free(aSR);
	free(aSS[0]); free(aSS);
	free(indxR); free(indxS);

    free(pair_scoreR); free(pair_scoreS);
    free(pair_scoreHy[0]); free(pair_scoreHy);
    
    free(pair_typeR[0]); free(pair_typeR);
    free(pair_typeS[0]); free(pair_typeS);
    for (i=0; i<num_align+2; i++) {
        free(pair_typeHy[i][0]); free(pair_typeHy[i]);
    }
    free(pair_typeHy);
    
	free(QbR); free(QbsR); free(QR); free(QmR);	free(QfR);
	free(QbS); free(QbsS); free(QS); free(QmS); free(QfS);

	for (j=0; j<20; j++) {
	  free(QRT[j][0]); free(QRT[j]);
	}
    
	for (j=0; j<20; j++) { 
	  free(QDT[j][0]); free(QDT[j]);
	}
    
	for (j=0; j<4; j++) {
        free(QTu[j][0]); free(QTu[j]); free(QTsu[j][0]); free(QTsu[j]);
        free(QTd[j][0]); free(QTd[j]); free(QTsd[j][0]); free(QTsd[j]);
        free(Bond[j][0]); free(Bond[j]);
	}
    
	free(QTe[0]); free(QTe); free(QTse[0]); free(QTse);
}

void encode_align(char **seqs_r, char **seqs_s)
{
    int i, j;
    
    for (i=1; i<=num_align; i++)
        for (j=1; j<=len_r; j++) {
            /* if (seqs_r[i][j] == '-') */
            /*     aSR[i][j] = 0; */
            /* else */
            aSR[i][j] = encode_char(toupper(seqs_r[i][j-1]));
        }
    for (i=1; i<=num_align; i++)
        for (j=1; j<=len_s; j++) {
            /* if (seqs_s[i][j] == '-') */
            /*     aSS[i][j] = 0; */
            /* else */
            aSS[i][j] = encode_char(toupper(seqs_s[i][j-1]));
        }
}

void make_pair_scores(char **seqs_r, char **seqs_s)
{
    int i, j, k, l, t;
    double score;
    
    int ham_dist[7][7] = {{0,0,0,0,0,0,0},  /* hamming distance between pairs */
                          {0,0,2,2,1,2,2},  /* CG */
                          {0,2,0,1,2,2,2},  /* GC */
                          {0,2,1,0,2,1,2},  /* GU */
                          {0,1,2,2,0,2,1},  /* UG */
                          {0,2,2,1,2,0,2},  /* AU */
                          {0,2,2,2,1,2,0}}; /* UA */

    for (j=1; j<=len_r; j++)
        for (i=(j>(TURN+2)?(j-TURN-2):1); i<=j; i++)
            pair_scoreR[indxR[j]+i] = NONE;
    for (j=1; j<=len_s; j++)
        for (i=(j>(TURN+2)?(j-TURN-2):1); i<=j; i++)
            pair_scoreS[indxS[j]+i] = NONE;

    //R
    for (l=TURN+4; l<=len_r; l++)     //l=j-i+1
        for (i=1; i<=(len_r-l+1); i++) {
            int n_pairs = 0;
            int pfreq[8] = {0,0,0,0,0,0,0,0};
            
            j = i+l-1;
            score = 0.0;

            for (k=1; k<=num_align; k++) {
                int type;

                if (seqs_r[k][i-1]=='.' || seqs_r[k][j-1]=='.')
                    type = 7;
                else
                    type = BP_pair[aSR[k][i]][aSR[k][j]];

                pfreq[type]++;
            }
            
            if ((pfreq[0]*2+pfreq[7]) >= num_align) {
                pair_scoreR[indxR[j]+i] = NONE;
                continue;
            }
            
            for (k=1; k<=6; k++)
                for (t=1; t<=6; t++)
                    score += pfreq[k]*ham_dist[k][t]*pfreq[t];
            if (num_align > 1)
                score /= num_align*(num_align-1)/2;

            for (t=1; t<=6; t++)
                n_pairs += pfreq[t];
            pair_scoreR[indxR[j]+i] = score-cv_fact*(1-n_pairs/num_align);
        }

    //S
    for (l=TURN+4; l<=len_s; l++)     //l=j-i+1
        for (i=1; i<=(len_s-l+1); i++) {
            int n_pairs = 0;
            int pfreq[8] = {0,0,0,0,0,0,0,0};
            
            j = i+l-1;
            score = 0.0;

            for (k=1; k<=num_align; k++) {
                int type;

                if (seqs_s[k][i-1]=='.' || seqs_s[k][j-1]=='.')
                    type = 7;
                else
                    type = BP_pair[aSS[k][i]][aSS[k][j]];

                pfreq[type]++;
            }
            
            if ((pfreq[0]*2+pfreq[7]) >= num_align) {
                pair_scoreS[indxS[j]+i] = NONE;
                continue;
            }
            
            for (k=1; k<=6; k++)
                for (t=1; t<=6; t++)
                    score += pfreq[k]*ham_dist[k][t]*pfreq[t];
            if (num_align > 1)
                score /= num_align*(num_align-1)/2;

            for (t=1; t<=6; t++)
                n_pairs += pfreq[t];
            pair_scoreS[indxS[j]+i] = score-cv_fact*(1-n_pairs/num_align);
        }

    //Hy
    for (i=1; i<=len_r; i++)
        for (j=1; j<=len_s; j++) {
            int n_pairs = 0;
            int pfreq[8] = {0,0,0,0,0,0,0,0};
           
            score = 0.0;
            for (k=1; k<=num_align; k++) {
                int type;

                if (seqs_r[k][i-1]=='.' || seqs_s[k][j-1]=='.')
                    type = 7;
                else
                    type = BP_pair[aSR[k][i]][aSS[k][j]];

                pfreq[type]++;
            }
            
            if ((pfreq[0]*2+pfreq[7]) >= num_align) {
                pair_scoreHy[i][j] = NONE;
                continue;
            }
            
            for (k=1; k<=6; k++)    /* ignore pairtype 7 (gap-gap) */
                for (t=1; t<=6; t++)
                    score += pfreq[k]*ham_dist[k][t]*pfreq[t];
            if (num_align > 1)
                score /= num_align*(num_align-1)/2;

            for (t=1; t<=6; t++)
                n_pairs += pfreq[t];
            pair_scoreHy[i][j] = score-cv_fact*(1-n_pairs/num_align);
        }

    return;
}

/**
 * make alignment pair types:
 * 0--no pair   1--CG   2--GC   3--GU      4--UG
 * 5--AU        6--UA   7--nostandards
 */
void make_align_ptypes(char **seqs_r, char **seqs_s, const char *structure)
{
    int i, j, k, l, type;

    make_pair_scores(seqs_r, seqs_s);
    
    //R
    for (l=1; l<=TURN+1; l++)
        for (i=1; i<=len_r-l+1; i++) {
            j = i+l-1;
            pair_typeR[0][indxR[j]+i] = NO_PAIR;
        }
    for (k=1; k<=num_align; k++)
        for (i=1; i<=len_r-TURN-1; i++)
            for (j=i+TURN+1; j<=len_r; j++) {
                if (pair_scoreR[indxR[j]+i] >= MINPSCORE) {
                    pair_typeR[0][indxR[j]+i] = PAIR;
                    if (seqs_r[k][i-1]=='.' || seqs_r[k][j-1]=='.')
                        type = 7;
                    else
                        type = BP_pair[aSR[k][i]][aSR[k][j]];
                
                    pair_typeR[k][indxR[j]+i] = type;
                }
                else
                    pair_typeR[0][indxR[j]+i] = NO_PAIR;
            }
    
    //S
    for (l=1; l<=TURN+1; l++)
        for (i=1; i<=len_s-l+1; i++) {
            j = i+l-1;
            pair_typeS[0][indxS[j]+i] = NO_PAIR;
        }
    for (k=1; k<=num_align; k++)
        for (i=1; i<=(len_s-TURN-1); i++)
            for (j=i+TURN+1; j<=len_s; j++) {
                if (pair_scoreS[indxS[j]+i] >= MINPSCORE) {
                    pair_typeS[0][indxS[j]+i] = PAIR;
                    if (seqs_s[k][i-1]=='.' || seqs_s[k][j-1]=='.')
                        type = 7;
                    else
                        type = BP_pair[aSS[k][i]][aSS[k][j]];

                    pair_typeS[k][indxS[j]+i] = type;
                }
                else
                    pair_typeS[0][indxS[j]+i] = NO_PAIR;
            }
    
    //Hybrid
    for (k=1; k<=num_align; k++)
        for (i=1; i<=len_r; i++)
            for (j=1; j<=len_s; j++) {
                if (pair_scoreHy[i][j] >= MINPSCORE) {
                    pair_typeHy[0][i][j] = PAIR;
                    if (seqs_r[k][i-1]=='.' || seqs_s[k][j-1]=='.')
                        type = 7;
                    else
                        type = BP_pair[aSR[k][i]][aSS[k][j]];
                    
                    pair_typeHy[k][i][j] = type;
                }
                else
                    pair_typeHy[0][i][j] = NO_PAIR;
            }

    /* ====================== constraint ========================= */
    if (constraint && structure) {
        int hx, *stack;
        int sh, *stack_hy;
        int js, t, l;
        int type, len;

        len = strlen(structure);
        stack = (int *) space(sizeof(int)*(len+1));
        stack_hy = (int *) space(sizeof(int)*(len+1));

        hx = 0;
        sh = 0;
        for(j=1; j<=len; j++) {
            if (j > len_r+1)
                js = j-len_r-1;
            
            switch (structure[j-1]) {
                case '-':           //it must pair one base in the same strand
                    BP[j] = -1;
                    
                    if (j <= len_r) {
                        for (l=1; l<=len_s; l++)
                            pair_typeHy[0][j][l] = NO_PAIR;
                    }
                    else {
                        for (l=1; l<=len_r; l++)
                            pair_typeHy[0][l][js] = NO_PAIR;
                    }
                    
                    break;
                case '^':           //it must pair one base in the other strand
                    BP[j] = -2;
                    
                    if (j <= len_r) {
                        for (l=1; l<j; l++)
                            pair_typeR[0][indxR[j]+l] = NO_PAIR;
                        for (l=j+1; l<=len_r; l++)
                            pair_typeR[0][indxR[l]+j] = NO_PAIR;
                    }
                    else {
                        for (l=1; l<js; l++)
                            pair_typeS[0][indxS[js]+l] = NO_PAIR;
                        for (l=js+1; l<=len_s; l++)
                            pair_typeS[0][indxS[l]+js] = NO_PAIR;
                    }
                    
                    break;
                case '<':           //pairs downstream
                    BP[j] = -3;
                    
                    if (j <= len_r) {
                        for (l=j+TURN+1; l<=len_r; l++)
                            pair_typeR[0][indxR[l]+j] = NO_PAIR;
                        for (l=1; l<=len_s; l++)
                            pair_typeHy[0][j][l] = NO_PAIR;
                    }
                    else {
                        for (l=1; l<(js-TURN); l++)
                            pair_typeS[0][indxS[js]+l] = NO_PAIR;
                        for (l=1; l<=len_r; l++)
                            pair_typeHy[0][l][js] = NO_PAIR;
                    }
                    
                    break;
                case '>':           //pairs upstream
                    BP[j] = -4;
                    
                    if (j <= len_r) {
                        for (l=1; l<j-TURN; l++)
                            pair_typeR[0][indxR[j]+l] = NO_PAIR;
                        for (l=1; l<=len_s; l++)
                            pair_typeHy[0][j][l] = NO_PAIR;
                    }
                    else {
                        for(l=js+TURN+1; l<=len_s; l++)
                            pair_typeS[0][indxS[l]+js] = NO_PAIR;
                        for (l=1; l<=len_r; l++)
                            pair_typeHy[0][l][js] = NO_PAIR;
                    }
                    
                    break;    
                case 'x':           //pair no base
                    BP[j] = -5;
                    
                    if (j <= len_r) {
                        for (l=1; l<j-TURN; l++)
                            pair_typeR[0][indxR[j]+l] = NO_PAIR;
                        for (l=j+TURN+1; l<=len_r; l++)
                            pair_typeR[0][indxR[l]+j] = NO_PAIR;
                        for (l=1; l<=len_s; l++)
                            pair_typeHy[0][j][l] = NO_PAIR;
                    }
                    else {
                        for (l=1; l<(js-TURN); l++)
                            pair_typeS[0][indxS[js]+l] = NO_PAIR;
                        for (l=js+TURN+1; l<=len_s; l++)
                            pair_typeS[0][indxS[l]+js] = NO_PAIR;
                        for (l=1; l<=len_r; l++)
                            pair_typeHy[0][l][js] = NO_PAIR;
                    }
                    
                    break;
                case '(':
                    if (j <= len_r)
                        stack[hx++] = j;
                    else
                        stack[hx++] = js;
                    
                    break;
                case ')':
                    if (hx <= 0) {
                        fprintf(stderr, "%s\n", structure);
                        fprintf(stderr, "unbalanced brackets in constraint\n");
                        exit(EXIT_FAILURE);
                    }
                    
                    i = stack[--hx];
                    if (j <= len_r)
                        type = pair_typeR[0][indxR[j]+i];
                    else
                        type = pair_typeS[0][indxS[js]+i];
                    if (type == NO_PAIR) {
                        fprintf(stderr, "%d and %d can not pair!\n", (j<=len_r) ? i : i+len_r+1, j);
                        exit(EXIT_FAILURE);
                    }
                    
                    if (j <= len_r) {
                        BP[i] = j;
                        BP[j] = i;
                        
                        /* don't allow pairs i<t<j<l */
                        for (l=j; l<=len_r; l++)
                            for (t=i+1; t<=j; t++)
                                pair_typeR[0][indxR[l]+t] = NO_PAIR;
                        for (l=1; l<i; l++)
                            pair_typeR[0][indxR[j]+l] = NO_PAIR;
                        /* don't allow pairs t<i<l<j */
                        for (l=i; l<j; l++)
                            for (t=1; t<=i; t++)
                                pair_typeR[0][indxR[l]+t] = NO_PAIR;
                        for (l=j+1; l<=len_r; l++)
                            pair_typeR[0][indxR[l]+i] = NO_PAIR;
                        
                        for (l=1; l<=len_s; l++) {
                            pair_typeHy[0][i][l] = NO_PAIR;
                            pair_typeHy[0][j][l] = NO_PAIR;
                        }
                    }
                    else {
                        BP[i+len_r+1] = js;
                        BP[j] = i;
                        
                        /* don't allow pairs i<t<j<l */
                        for (l=js; l<=len_s; l++)
                            for (t=i+1; t<=js; t++)
                                pair_typeS[0][indxS[l]+t] = NO_PAIR;
                        for (l=1; l<i; l++)
                            pair_typeS[0][indxS[j]+l] = NO_PAIR;
                        
                        /* don't allow pairs t<i<l<j */
                        for (l=i; l<js; l++)
                            for (t=1; t<=i; t++)
                                pair_typeS[0][indxS[l]+t] = NO_PAIR;
                        for (l=js+1; l<=len_s; l++)
                            pair_typeS[0][indxS[l]+i] = NO_PAIR;
                        
                        for (l=1; l<=len_r; l++) {
                            pair_typeHy[0][l][i] = NO_PAIR;
                            pair_typeHy[0][l][js] = NO_PAIR;
                        }
                    }
                    
                    break;
                case '[':
                    if (j > len_r) {
                        fprintf(stderr, "bad constraint on hybrid!\n");
                        exit(EXIT_FAILURE);
                    }
                    stack_hy[sh++] = j;
                    
                    break;
                case ']':
                    if (j < len_r+2) {
                        fprintf(stderr, "bad constraint on hybrid!\n");
                        exit(EXIT_FAILURE);
                    }
                    if (sh <= 0) {
                        fprintf(stderr, "unbalanced brackets on hybrid!\n");
                        exit(EXIT_FAILURE);
                    }
                    
                    i = stack_hy[--sh];
                    if (pair_typeHy[0][i][js] == NO_PAIR) {
                        fprintf(stderr, "%d and %d can not pair!\n", i, j);
                        exit(EXIT_FAILURE);
                    }
                    
                    BP[i] = j;
                    BP[j] = i;
                    
                    for (l=1; l<i; l++)
                        pair_typeR[0][indxR[i]+l] = NO_PAIR;
                    for (l=i+1; l<=len_r; l++)
                        pair_typeR[0][indxR[l]+i] = NO_PAIR;
                    
                    for (l=1; l<js; l++)
                        pair_typeS[0][indxS[js]+l] = NO_PAIR;
                    for (l=js+1; l<=len_s; l++)
                        pair_typeS[0][indxS[l]+js] = NO_PAIR;
                    
                    for (l=1; l<i; l++)
                        pair_typeHy[0][l][js] = NO_PAIR;
                    for (l=i+1; l<=len_r; l++)
                        pair_typeHy[0][l][js] = NO_PAIR;
                    for (l=1; l<js; l++)
                        pair_typeHy[0][i][l] = NO_PAIR;
                    for (l=js+1; l<=len_s; l++)
                        pair_typeHy[0][i][l] = NO_PAIR;
                    
                    break;
                default:
                    break;
            }
        }
        
        if (hx!=0 || sh!=0) {
            fprintf(stderr, "%s\n", structure);
            fprintf(stderr, "unbalanced brackets in constraint string\n");
            exit(EXIT_FAILURE);
        }
        
        free(stack);
        free(stack_hy);
    }   //end if (constraint && structure)

    return;
}

void calc_secondary_pfunc(FILE *fp, char **seqs, int **S, double *Q, double *Qb, double *Qbs, 
                          double *Qm, double *Qf, int *indx, int **pair_type, int *bp)
{
	int i, j, d, e, l, k, len, ij, de;
    int tmp_type, tmp_type_2;
	int type, type_2, type_i, type_2i;
	int G = 0, *stack_G;
    int bonus = 0;
    
    stack_G = (int *) space(sizeof(int)*(num_align+1));
    
    len = strlen(seqs[1]);
    //initialization
    for (j=1; j<=len; j++)
        for (i=(j>(TURN+2)?(j-TURN-2):1); i<=j; i++)
            Q[indx[j]+i] = 1.0;
        
    for (l=TURN+4; l<=len; l++)
        for (i=1; i<=(len-l+1); i++) {
            j = i+l-1;
            ij = indx[j]+i;
            bonus = 0;

            type = pair_type[0][ij];
            type_i = pair_type[0][indx[j-1]+i+1];
            
            //Qbs recursion
            if (type==PAIR && type_i==NO_PAIR) {
                if (constraint) {
                    //enforcing structure constraint
                    if (bp[i]==j || bp[i]==-1 || bp[i]==-4)
                        bonus -= PF_BONUS;
                    if (bp[j]==-1 || bp[j]==-3)
                        bonus -= PF_BONUS;
                }
            
                //hairpin loop
                G = 0;
                for (k=1; k<=num_align; k++) {
                    tmp_type = pair_type[k][ij];
                    
                    if (tmp_type == 0)
                        tmp_type = 7;
                    G += calc_hairpin_energy(seqs[k]+i-1, j-i-1, tmp_type, S[k][i+1], S[k][j-1]);
                }
                Qbs[ij] = EXP(((double) G)/num_align);
                
                for (d=i+1; d<j-TURN-3; d++)
                    for (e=d+TURN+3; e<j; e++) {
                        de = indx[e]+d;
                        type_2 = pair_type[0][de];
                        type_2i = pair_type[0][indx[e-1]+d+1];
                        
                        if (type_2==NO_PAIR || type_2i==NO_PAIR)
                            continue;
                        if (d==i+1 && e==j-1)
                            continue;
                        
                        //interior loop
                        G = 0;
                        for (k=1; k<=num_align; k++) {
                            tmp_type = pair_type[k][ij];
                            tmp_type_2 = pair_type[k][de];
                            
                            if (tmp_type == 0)
                                tmp_type = 7;
                            if (tmp_type_2 == 0)
                                tmp_type_2 = 7;
                            G += calc_interior_energy(d-i-1, j-e-1, tmp_type, rtype[tmp_type_2],
                                                      S[k][i+1], S[k][j-1], S[k][d-1], S[k][e+1]);
                        }
                        Qbs[ij] += Qb[de]*EXP(((double) G)/num_align);
                        
                        //multi-loop
                        if ((d-i-2) >= (TURN+3)) {
                            G = 0;
                            for (k=1; k<=num_align; k++) {
                                G += P->MLclosing + P->MLintern[pair_type[k][ij]]
                                    + P->MLintern[pair_type[k][de]] + (j-e-1)*P->MLbase;
                            }
                            Qbs[ij] += Qm[indx[d-1]+i+1] * Qb[de] * EXP(((double) G)/num_align);
                        }
                    }

                if (constraint)
                    Qbs[ij] *= EXP(bonus);  //(Qbs[ij]+PF_BASE)*bonus;
            }
            else
                Qbs[ij] = 0.0;
            
            //Qb recursion
            if (type==PAIR && type_i==PAIR) {
                if (constraint) {
                    if (bp[i]==j || bp[i]==-1 || bp[i]==-4)
                        bonus -= PF_BONUS;
                    if (bp[j]==-1 || bp[j]==-3)
                        bonus -= PF_BONUS;
                    if (bp[i+1]==j-1 || bp[i+1]==-1 || bp[i+1]==-4)
                        bonus -= PF_BONUS;
                    if (bp[j-1]==-1 || bp[j-1]==-3)
                        bonus -= PF_BONUS;
                }

                for (k=1; k<=num_align; k++) {
                    tmp_type = pair_type[k][ij];
                    tmp_type_2 = pair_type[k][indx[j-1]+i+1];
                    
                    if (tmp_type == 0)
                        tmp_type = 7;
                    if (tmp_type_2 == 0)
                        tmp_type_2 = 7;
                    stack_G[k] = P->stack[tmp_type][rtype[tmp_type_2]];
                }
                
                //hairpin loop
                G = 0;
                for (k=1; k<=num_align; k++) {
                    tmp_type = pair_type[k][indx[j-1]+i+1];
                    
                    if (tmp_type == 0)
                        tmp_type = 7;
                    G += calc_hairpin_energy(seqs[k]+i, j-i-3, tmp_type, S[k][i+2], S[k][j-2])+stack_G[k];
                }
                Qb[ij] = EXP(((double) G)/num_align);
                
                //form a stack
                G = 0;
                for (k=1; k<=num_align; k++) {
                    tmp_type = pair_type[k][indx[j-1]+i+1];
                    tmp_type_2 = pair_type[k][indx[j-2]+i+2];
                    
                    if (tmp_type == 0)
                        tmp_type = 7;
                    if (tmp_type_2 == 0)
                        tmp_type_2 = 7;
                    G += stack_G[k] + P->stack[tmp_type][rtype[tmp_type_2]];
                }
                Qb[ij] += EXP(((double) G)/num_align) * Qbs[indx[j-2]+i+2];
                
                for (d=i+2; d<j-TURN-4; d++)
                    for (e=d+TURN+3; e<j-1; e++) {
                        de = indx[e]+d;
                        type_2 = pair_type[0][de];
                        type_2i = pair_type[0][indx[e-1]+d+1];
                        
                        if (type_2==NO_PAIR || type_2i==NO_PAIR)
                            continue;
                       
                        //interior loop
                        G = 0;
                        for (k=1; k<=num_align; k++) {
                            tmp_type = pair_type[k][indx[j-1]+i+1];
                            tmp_type_2 = pair_type[k][de];

                            if (tmp_type == 0)
                                tmp_type = 7;
                            if (tmp_type_2 == 0)
                                tmp_type_2 = 7;
                            G += calc_interior_energy(d-i-2, j-e-2, tmp_type, rtype[tmp_type_2],
                                                      S[k][i+2], S[k][j-2], S[k][d-1], S[k][e+1]) + stack_G[k];
                        }
                        Qb[ij] += Qb[de] * EXP(((double) G)/num_align);
                        
                        //multi-loop
                        if ((d-i-3) >= (TURN+3)) {
                            G = 0;
                            for (k=1; k<=num_align; k++) {
                                G += stack_G[k] + P->MLclosing + P->MLintern[pair_type[k][indx[j-1]+i+1]]
                                    + P->MLintern[pair_type[k][de]] + (j-e-2)*P->MLbase;
                            }
                            Qb[ij] += Qm[indx[d-1]+i+2] * Qb[de] * EXP(((double) G)/num_align);
                        }
                    }

                if (constraint)
                    Qb[ij] *= EXP(bonus);   //(Qb[ij]+PF_BASE)*bonus;
            }
            else
                Qb[ij] = 0.0;
            /* if (Qb[ij] != 0.0) */
            /*     fprintf(fp, "Qb[%d,%d]=%f\n", i, j, Qb[ij]); */
                
            //Q, Qm and Qf recursion
            Q[ij] = 1.0;
            for (d=i; d<j-TURN-2; d++)
                for (e=d+TURN+3; e<=j; e++) {
                    de = indx[e]+d;
                    type_2 = pair_type[0][de];
                    type_2i = pair_type[0][indx[e-1]+d+1];
                    
                    if (type_2==NO_PAIR || type_2i==NO_PAIR)
                        continue;

                    //Q
                    if (d == i)
                        Q[ij] += Qb[de];
                    else
                        Q[ij] += Q[indx[d-1]+i] * Qb[de];

                    //Qm
                    G = 0;
                    for (k=1; k<=num_align; k++)
                        G += (j-e)*P->MLbase+P->MLintern[pair_type[k][de]];
                    if (d == i)
                        Qm[ij] += Qb[de]*EXP(((double) G)/num_align);
                    else
                        Qm[ij] += (Qm[indx[d-1]+i]+EXP((d-i)*P->MLbase))*Qb[de]*EXP(((double) G)/num_align);

                    //Qf
                    if (d == i)
                        Qf[ij] += Qb[de]*EXP(BETA2+(j-e)*BETA3);
                    else
                        Qf[ij] += (Qf[indx[d-1]+i]+EXP((d-i)*BETA3))*Qb[de]*EXP(BETA2+(j-e)*BETA3);
                }
            /* if (Q[ij] > 1.0) */
            /*     fprintf(fp, "Q[%d,%d]=%f\n", i, j, Q[ij]); */
        }

    fprintf(fp, "PF of Sec: %16.16e\n", Q[indx[len]+1]);
    
    free(stack_G);
}

double get_gap_pfunc(int i, int j, int k1, int k2, int sig, double *Qm, double *Qf, int *indx)
{
	// sig=1:  multi 1  
	// sig=2:  multi 2  
	// sig=3:  kissing 
	double res = 0.0;
	double left, right;
    
	if (sig == 1) {       //there must exist arcs in the two gaps
        if (k1-i-1<TURN+2 && j-k2-1<TURN+2)
            return 0.0;
        else {
            if (k1-i-1<TURN+2 && j-k2-1>=TURN+2) {
                left = EXP((k1-i-1)*P->MLbase);
                right = Qm[indx[j-1]+k2+1];
                
                return left*right;
            }
            else if (k1-i-1>=TURN+2 && j-k2-1<TURN+2) {
                left = Qm[indx[k1-1]+i+1];
                right = EXP((j-k2-1)*P->MLbase);
                
                return left*right;
            }
            else if (k1-i-1>=TURN+2 && j-k2-1>=TURN+2) {
                res += Qm[indx[k1-1]+i+1]*Qm[indx[j-1]+k2+1]
                    + Qm[indx[k1-1]+i+1]*EXP((j-k2-1)*P->MLbase)
                    + Qm[indx[j-1]+k2+1]*EXP((k1-i-1)*P->MLbase);
                
                return res;
            }
        }
	}
    else if (sig == 2) {    //arbitrary secondary structures in the two gaps
        if (k1-i-1 < TURN+2) 
            left = EXP((k1-i-1)*P->MLbase);
        else 
            left = Qm[indx[k1-1]+i+1] + EXP((k1-i-1)*P->MLbase);
        if (j-k2-1 < TURN+2)
            right = EXP((j-k2-1)*P->MLbase);
        else 
            right = Qm[indx[j-1]+k2+1] + EXP((j-k2-1)*P->MLbase);
        
        return left*right;
	}
    else if (sig == 3) {    //contained in a kissing loop
        if (k1-i-1 < TURN+2) 
            left = EXP((k1-i-1)*BETA3);
        else 
            left = Qf[indx[k1-1]+i+1] + EXP((k1-i-1)*BETA3);
        if (j-k2-1 < TURN+2)
            right = EXP((j-k2-1)*BETA3);
        else 
            right = Qf[indx[j-1]+k2+1] + EXP((j-k2-1)*BETA3);
        
        return left*right;
	}

    return res;
}

double calc_Tu_pfunc(double **QTu, double **QTsu, double **QDTm, double **QDTka, double **QDTkb, 
                int ir, int jr, int is, int js, int sig)
{
  	int k, k1, k2, irjr, isjs, irjr_i, k1k2;
    int type1, type2;
    int type, type_2, type_2i;
	double res = 0.0;
	int G;
    int bonus = 0;
    
	irjr = indxR[jr]+ir;
	isjs = indxS[js]+is;
    irjr_i = indxR[jr-1]+ir+1;
    
    type = pair_typeR[0][irjr];

    if (sig == 1) {   // PF of Tu
        int type_i, *stack_G;
        
        stack_G = (int *) space(sizeof(int) * (num_align+1));
        type_i = pair_typeR[0][irjr_i];
        
        if (type==NO_PAIR || type_i==NO_PAIR)
            return 0.0;

        if (constraint) {
            if (BP[ir]==jr || BP[ir]==-1 || BP[ir]==-4)
                bonus -= PF_BONUS;
            if (BP[jr]==-1 || BP[jr]==-3)
                bonus -= PF_BONUS;
            if (BP[ir+1]==jr-1 || BP[ir+1]==-1 || BP[ir+1]==-4)
                bonus -= PF_BONUS;
            if (BP[jr-1]==-1 || BP[jr-1]==-3)
                bonus -= PF_BONUS;
        }

        for (k=1; k<=num_align; k++) {
            type1 = pair_typeR[k][irjr];
            type2 = pair_typeR[k][irjr_i];

            if (type1 == 0)
                type1 = 7;
            if (type2 == 0)
                type2 = 7;
            stack_G[k] = P->stack[type1][rtype[type2]];
        }
        
        //Tsu is contained
        G = 0;
        for (k=1; k<=num_align; k++) {
            type1 = pair_typeR[k][irjr_i];
            type2 = pair_typeR[k][indxR[jr-2]+ir+2];

            if (type1 == 0)
                type1 = 7;
            if (type2 == 0)
                type2 = 7;
            G += stack_G[k]+P->stack[type1][rtype[type2]];
        }
        res += QTsu[indxR[jr-2]+ir+2][isjs]*EXP(((double) G)/num_align);
         
        for (k1=ir+2; k1<jr-2; k1++)
            for (k2=k1+1; k2<jr-1; k2++) {
                k1k2 = indxR[k2]+k1;
                
                type_2 = pair_typeR[0][k1k2];
                if ((k2-k1-2) >= TURN+1)
                    type_2i = pair_typeR[0][indxR[k2-1]+k1+1];
                else
                    type_2i = NO_PAIR;    //at least one sequence has no pair 

                if (type_2==PAIR && type_2i==PAIR) {
                    //form an interior loop
                    G = 0;
                    for (k=1; k<=num_align; k++) {
                        type1 = pair_typeR[k][irjr_i];
                        type2 = pair_typeR[k][k1k2];

                        if (type1 == 0)
                            type1 = 7;
                        if (type2 == 0)
                            type2 = 7;
                        G += calc_interior_energy(k1-ir-2, jr-k2-2, type1, rtype[type2],
                                                  aSR[k][ir+2], aSR[k][jr-2], aSR[k][k1-1], aSR[k][k2+1])+stack_G[k];
                    }
                    res += QTu[k1k2][isjs] * EXP(((double) G)/num_align);
                    
                    //form a multi-loop
                    G =  0;
                    for (k=1; k<=num_align; k++) {
                        G += P->MLclosing + P->MLintern[pair_typeR[k][irjr_i]]
                            + P->MLintern[pair_typeR[k][k1k2]] + stack_G[k];    // Tu is contained
                    }
                    res += get_gap_pfunc(ir+1,jr-1,k1,k2,1,QmR,QfR,indxR)*QTu[k1k2][isjs]*EXP(((double) G)/num_align);
                }

                // dts or Td is contained
                G = 0;
                for (k=1; k<=num_align; k++)
                    G += P->MLclosing + P->MLintern[pair_typeR[k][irjr_i]] + stack_G[k];
                res += get_gap_pfunc(ir+1,jr-1,k1,k2,2,QmR,QfR,indxR)*QDTm[k1k2][isjs]*EXP(((double) G)/num_align);
                
                //form a kissing loop
                G = 0;
                for (k=1; k<=num_align; k++)
                    G += BETA1+BETA2+stack_G[k];
                if (QDTkb != NULL) 
                    res += get_gap_pfunc(ir+1,jr-1,k1,k2,3,QmR,QfR,indxR)
                        * (QDTka[k1k2][isjs]+QDTkb[k1k2][isjs])*EXP(((double) G)/num_align);
                else
                    res += get_gap_pfunc(ir+1,jr-1,k1,k2,3,QmR,QfR,indxR)*QDTka[k1k2][isjs]
                        * EXP(((double) G)/num_align);
            }
        
        free(stack_G);
    }
    else if (sig == 2) {      //PF of Tsu
        if (type == NO_PAIR)
            return 0.0;
        
        if (constraint) {
            if (BP[ir]==jr || BP[ir]==-1 || BP[ir]==-4)
                bonus -= PF_BONUS;
            if (BP[jr]==-1 || BP[jr]==-3)
                bonus -= PF_BONUS;
        }
        
        for (k1=ir+1; k1<jr-1; k1++)
            for (k2=k1+1; k2<jr; k2++) {
                k1k2 = indxR[k2]+k1;
                
                type_2 = pair_typeR[0][k1k2];
                if ((k2-k1-2) >= TURN+1)
                    type_2i = pair_typeR[0][indxR[k2-1]+k1+1];
                else
                    type_2i = NO_PAIR;
                
                if (type_2==PAIR && type_2i==PAIR && (k1!=ir+1 || k2!=jr-1)) {
                    // form an interior loop
                    G = 0;
                    for (k=1; k<=num_align; k++) {
                        type1 = pair_typeR[k][irjr];
                        type2 = pair_typeR[k][k1k2];

                        if (type1 == 0)
                            type1 = 7;
                        if (type2 == 0)
                            type2 = 7;
                        G += calc_interior_energy(k1-ir-1, jr-k2-1, type1, rtype[type2],
                                                  aSR[k][ir+1], aSR[k][jr-1], aSR[k][k1-1], aSR[k][k2+1]);
                    }
                    res += QTu[k1k2][isjs] * EXP(((double) G)/num_align);
                    
                    // form a multi-loop
                    G = 0;
                    for (k=1; k<=num_align; k++) {
                        G += P->MLclosing + P->MLintern[pair_typeR[k][irjr]]
                            + P->MLintern[pair_typeR[k][k1k2]];   //Tu is contained
                    }
                    res += get_gap_pfunc(ir,jr,k1,k2,1,QmR,QfR,indxR) * QTu[k1k2][isjs] * EXP(((double) G)/num_align);
                }

                G = 0;
                for (k=1; k<=num_align; k++)
                    G += P->MLclosing + P->MLintern[pair_typeR[k][irjr]];        // dts or Td is contained
                res += get_gap_pfunc(ir,jr,k1,k2,2,QmR,QfR,indxR) * QDTm[k1k2][isjs] * EXP(((double) G)/num_align);
                
                // form a kissing loop
                G = BETA1+BETA2;
                if (QDTkb)
                    res += get_gap_pfunc(ir,jr,k1,k2,3,QmR,QfR,indxR)
                        * (QDTka[k1k2][isjs]+QDTkb[k1k2][isjs]) * EXP(G);
                else
                    res += get_gap_pfunc(ir,jr,k1,k2,3,QmR,QfR,indxR) * QDTka[k1k2][isjs] * EXP(G);
            }
    }

    if (constraint)
        res *= EXP(bonus);

    return res;
}

double calc_Td_pfunc(double **QTd, double **QTsd, double **QDTm, double **QDTka, double **QDTkb,
                int ir, int jr, int is, int js, int sig)
{
    //sig=2: PF of Tsd
	int k1, k2, irjr, isjs, k, isjs_i, k1k2;
    int type1, type2;
    int type, type_2, type_2i;
	double res = 0.0;
	int G;
    int bonus = 0;
    
	irjr = indxR[jr]+ir;
	isjs = indxS[js]+is;
    isjs_i = indxS[js-1]+is+1;

    type = pair_typeS[0][isjs];
    
    if (sig == 1) {         /* partition function of Td */
        int type_i, *stack_G;

        stack_G = (int *) space(sizeof(int) * (num_align+1));
        type_i = pair_typeS[0][isjs_i];
        
        if (type==NO_PAIR || type_i==NO_PAIR)
            return 0.0;

        if (constraint) {
            if (BP[BP[0]+is+1]==js || BP[BP[0]+is+1]==-1 || BP[BP[0]+is+1]==-3)
                bonus -= PF_BONUS;
            if (BP[BP[0]+js+1]==-1 || BP[BP[0]+js+1]==-4)
                bonus -= PF_BONUS;
            if (BP[BP[0]+is+2]==js-1 || BP[BP[0]+is+2]==-1 || BP[BP[0]+is+2]==-3)
                bonus -= PF_BONUS;
            if (BP[BP[0]+js]==-1 || BP[BP[0]+js]==-4)
                bonus -= PF_BONUS;
        }

        for (k=1; k<=num_align; k++) {
            type1 = pair_typeS[k][isjs];
            type2 = pair_typeS[k][isjs_i];

            if (type1 == 0)
                type1 = 7;
            if (type2 == 0)
                type2 = 7;
            stack_G[k] = P->stack[rtype[type1]][type2];
        }
        //Tsd is contained
        G = 0;
        for (k=1; k<=num_align; k++) {
            type1 = pair_typeS[k][isjs_i];
            type2 = pair_typeS[k][indxS[js-2]+is+2];

            if (type1 == 0)
                type1 = 7;
            if (type2 == 0)
                type2 = 7;
            G += stack_G[k] + P->stack[rtype[type1]][type2];
        }
        res += QTsd[irjr][indxS[js-2]+is+2] * EXP(((double) G)/num_align);

        for (k1=is+2; k1<js-2; k1++)
            for (k2=k1+1; k2<js-1; k2++) {
                k1k2 = indxS[k2]+k1;
                
                type_2 = pair_typeS[0][k1k2];
                if ((k2-k1-2) >= TURN+1)
                    type_2i = pair_typeS[0][indxS[k2-1]+k1+1];
                else
                    type_2i = NO_PAIR;
                
                if (type_2==PAIR && type_2i==PAIR) {
                    //form an interior loop
                    G = 0;
                    for (k=1; k<=num_align; k++) {
                        type1 = pair_typeS[k][isjs_i];
                        type2 = pair_typeS[k][k1k2];

                        if (type1 == 0)
                            type1 = 7;
                        if (type2 == 0)
                            type2 = 7;
                        G += calc_interior_energy(js-k2-2, k1-is-2, rtype[type1], type2,
                                                  aSS[k][js-2], aSS[k][is+2], aSS[k][k2+1], aSS[k][k1-1])
                            + stack_G[k];
                    }
                    res += QTd[irjr][k1k2] * EXP(((double) G)/num_align);
                    
                    //form a multi-loop
                    G = 0;
                    for (k=1; k<=num_align; k++) {
                        G += P->MLclosing + P->MLintern[rtype[pair_typeS[k][isjs_i]]]
                            + P->MLintern[rtype[pair_typeS[k][k1k2]]] + stack_G[k];      //Td is contained
                    }
                    res += get_gap_pfunc(is+1,js-1,k1,k2,1,QmS,QfS,indxS) * QTd[irjr][k1k2]
                        * EXP(((double) G)/num_align);
                }
                
                G = 0;
                for (k=1; k<=num_align; k++)
                    G += P->MLclosing + P->MLintern[rtype[pair_typeS[k][isjs_i]]] + stack_G[k];       //dts is contained
                res += get_gap_pfunc(is+1,js-1,k1,k2,2,QmS,QfS,indxS) * QDTm[irjr][k1k2] * EXP(((double) G)/num_align);
                
                //form a kissing loop
                G = 0;
                for (k=1; k<=num_align; k++)
                    G += BETA1 + BETA2 + stack_G[k];
                if (QDTkb != NULL) 
                    res += get_gap_pfunc(is+1,js-1,k1,k2,3,QmS,QfS,indxS)
                        * (QDTka[irjr][k1k2]+QDTkb[irjr][k1k2]) * EXP(((double) G)/num_align);
                else
                    res += get_gap_pfunc(is+1,js-1,k1,k2,3,QmS,QfS,indxS) * QDTka[irjr][k1k2]
                        * EXP(((double) G)/num_align);
            }
        
        free(stack_G);
    }
    else if (sig == 2) {        /* partition function of Tsd */
        if (type == NO_PAIR)
            return 0.0;

        if (constraint) {
            if (BP[BP[0]+is+1]==js || BP[BP[0]+is+1]==-1 || BP[BP[0]+is+1]==-3)
                bonus -= PF_BONUS;
            if (BP[BP[0]+js+1]==-1 || BP[BP[0]+js+1]==-4)
                bonus -= PF_BONUS;
        }
        
        for (k1=is+1; k1<js-1; k1++)
            for (k2=k1+1; k2<js; k2++) {
                k1k2 = indxS[k2]+k1;
                
                type_2 = pair_typeS[0][k1k2];
                if ((k2-k1-2) >= TURN+1)
                    type_2i = pair_typeS[0][indxS[k2-1]+k1+1];
                else
                    type_2i = NO_PAIR;    //at least one sequence has no pair
                
                if (type_2==PAIR && type_2i==PAIR && (k1!=is+1 || k2!=js-1)) {
                    //form an interior loop
                    G = 0;
                    for(k=1; k<=num_align; k++) {
                        type1 = pair_typeS[k][isjs];
                        type2 = pair_typeS[k][k1k2];

                        if (type1 == 0)
                            type1 = 7;
                        if (type2 == 0)
                            type2 = 7;
                        G += calc_interior_energy(js-k2-1, k1-is-1, rtype[type1], type2,
                                                  aSS[k][js-1], aSS[k][is+1], aSS[k][k2+1], aSS[k][k1-1]);
                    }
                    res += QTd[irjr][k1k2] * EXP(((double) G)/num_align);
                    
                    //form a multi-loop
                    G = 0;
                    for (k=1; k<=num_align; k++) {
                        G += P->MLclosing + P->MLintern[rtype[pair_typeS[k][isjs]]]
                            + P->MLintern[rtype[pair_typeS[k][k1k2]]];      //Tu is contained
                    }
                    res += get_gap_pfunc(is,js,k1,k2,1,QmS,QfS,indxS) * QTd[irjr][k1k2] * EXP(((double) G)/num_align);
                }

                G = 0;
                for (k=1; k<=num_align; k++)
                    G += P->MLclosing + P->MLintern[rtype[pair_typeS[k][isjs]]];     //dts is contained
                res += get_gap_pfunc(is,js,k1,k2,2,QmS,QfS,indxS) * QDTm[irjr][k1k2] * EXP(((double) G)/num_align);
                
                //form a kissing loop
                G = BETA1+BETA2;
                if (QDTkb != NULL) 
                    res += get_gap_pfunc(is,js,k1,k2,3,QmS,QfS,indxS)
                        * (QDTka[irjr][k1k2]+QDTkb[irjr][k1k2]) * EXP(G);
                else
                    res += get_gap_pfunc(is,js,k1,k2,3,QmS,QfS,indxS) * QDTka[irjr][k1k2] * EXP(G);
            }
    }

    if (constraint)
        res *= EXP(bonus);

    return res;
}

void print_block_pf()
{
    int i, j, r, s, t;

    //R
    for (j=len_r-1; j>=TURN+1; j--)
        for (i=1; i<=len_r-j; i++) {
            if (QR[indxR[i+j]+i] != 0)
                fprintf(stdout, "QR[%d,%d]=%8.8e\n", i, i+j, QR[indxR[i+j]+i]);

            if (QbR[indxR[i+j]+i] != 0)
                fprintf(stdout, "QbR[%d,%d]=%8.8e\n", i, i+j, QbR[indxR[i+j]+i]);

            if (QbsR[indxR[i+j]+i] != 0)
                fprintf(stdout, "QbsR[%d,%d]=%8.8e\n", i, i+j, QbsR[indxR[i+j]+i]);

            if (QmR[indxR[i+j]+i] != 0)
                fprintf(stdout, "QR[%d,%d]=%8.8e\n", i, i+j, QmR[indxR[i+j]+i]);

            if (QfR[indxR[i+j]+i] != 0)
                fprintf(stdout, "QfR[%d,%d]=%8.8e\n", i, i+j, QfR[indxR[i+j]+i]);
        }
    //S
    for (j=len_s-1; j>=TURN+1; j--)
        for (i=1; i<=len_s-j; i++) {
            if (QS[indxS[i+j]+i] != 0)
                fprintf(stdout, "QS[%d,%d]=%8.8e\n", i, i+j, QS[indxS[i+j]+i]);

            if (QbS[indxS[i+j]+i] != 0)
                fprintf(stdout, "QbS[%d,%d]=%8.8e\n", i, i+j, QbS[indxS[i+j]+i]);

            if (QbsS[indxS[i+j]+i] != 0)
                fprintf(stdout, "QbsS[%d,%d]=%8.8e\n", i, i+j, QbsS[indxS[i+j]+i]);

            if (QmS[indxS[i+j]+i] != 0)
                fprintf(stdout, "QS[%d,%d]=%8.8e\n", i, i+j, QmS[indxS[i+j]+i]);

            if (QfS[indxS[i+j]+i] != 0)
                fprintf(stdout, "QfS[%d,%d]=%8.8e\n", i, i+j, QfS[indxS[i+j]+i]);
        }

    /* RT & DT */
    for (j=len_r-1; j>=0; j--)
          for (i=1; i<=len_r-j; i++) 
              for (s=len_s-1; s>=0; s--)
                  for (r=1; r<=len_s-s; r++)
                      for (t=0; t<20; t++) {
                          if (QRT[t][indxR[i+j]+i][indxS[r+s]+r] != 0)
                              fprintf(stdout, "QRT[%d][%d,%d][%d,%d]= %8.8e\n", t, i, i+j, r, r+s,
                                      QRT[t][indxR[i+j]+i][indxS[r+s]+r]);

                          if (QDT[t][indxR[i+j]+i][indxS[r+s]+r] != 0)
                              fprintf(stdout, "QDT[%d][%d,%d][%d,%d]= %8.8e\n", t, i, i+j, r, r+s,
                                      QDT[t][indxR[i+j]+i][indxS[r+s]+r]);
                      }

    /* Tu & Td & Te*/
    for (j=len_r-1; j>=0; j--)
          for (i=1; i<=len_r-j; i++) 
              for (s=len_s-1; s>=0; s--)
                  for (r=1; r<=len_s-s; r++) {
                      for (t=0; t<4; t++) {
                          if (QTu[t][indxR[i+j]+i][indxS[r+s]+r] != 0)
                              fprintf(stdout, "QTu[%d][%d,%d][%d,%d]= %8.8e\n", t, i, i+j, r, r+s,
                                      QTu[t][indxR[i+j]+i][indxS[r+s]+r]);

                          if (QTsu[t][indxR[i+j]+i][indxS[r+s]+r] != 0)
                              fprintf(stdout, "QTsu[%d][%d,%d][%d,%d]= %8.8e\n", t, i, i+j, r, r+s,
                                      QTsu[t][indxR[i+j]+i][indxS[r+s]+r]);

                          if (QTd[t][indxR[i+j]+i][indxS[r+s]+r] != 0)
                              fprintf(stdout, "QTd[%d][%d,%d][%d,%d]= %8.8e\n", t, i, i+j, r, r+s,
                                      QTd[t][indxR[i+j]+i][indxS[r+s]+r]);

                          if (QTsd[t][indxR[i+j]+i][indxS[r+s]+r] != 0)
                              fprintf(stdout, "QTsd[%d][%d,%d][%d,%d]= %8.8e\n", t, i, i+j, r, r+s,
                                      QTsd[t][indxR[i+j]+i][indxS[r+s]+r]);
                      }

                      if (QTe[indxR[i+j]+i][indxS[r+s]+r] != 0)
                              fprintf(stdout, "QTe[%d,%d][%d,%d]= %8.8e\n", i, i+j, r, r+s,
                                      QTe[indxR[i+j]+i][indxS[r+s]+r]);

                      if (QTse[indxR[i+j]+i][indxS[r+s]+r] != 0)
                              fprintf(stdout, "QTse[%d,%d][%d,%d]= %8.8e\n", i, i+j, r, r+s,
                                      QTse[indxR[i+j]+i][indxS[r+s]+r]);

                      for (t=0; t<4; t++) {
                          if (Bond[t][indxR[i+j]+i][indxS[r+s]+r] != 0)
                              fprintf(stdout, "Bond[%d][%d,%d][%d,%d]= %8.8e\n", t, i, i+j, r, r+s,
                                      Bond[t][indxR[i+j]+i][indxS[r+s]+r]);
                      }
                      
                  }

    return;
}

void calc_align_pfunc(FILE *fp, char **seqs_r, char **seqs_s_53, char *cons_stru)
{
	char *ptr, **seqs_s_35;
    int k, len;
	int ir, jr, is, js, k1, k2, gr, gs;
	int irjr, isjs;
    int bonus_hy;
	double Tight;
	int G, GMR, GMS, GFR, GFS;
    int type1, type2;
	int typeR, typeS, typeI, type_I2, type_Ik, type_kR, type_kS, typeR_i, typeS_i;

    for (k=1; k<=num_align; k++)
        fprintf(fp, "%s|%s\n", seqs_r[k], seqs_s_53[k]);
    
	seqs_s_35 = (char **) space((num_align+1) * sizeof(char *));
    ptr = (char *) space((num_align+1) * (len_s+2) * sizeof(char));
    for (k=0; k<=num_align; k++) {
        seqs_s_35[k] = ptr;
        ptr += len_s+2;
    }
	reverse_align(seqs_s_53, seqs_s_35);

    if (constraint) {
        BP = (int *) space(sizeof(int)*(len_r+len_s+2));
        BP[0] = len_r;
        BP[len_r+1] = -10;   //separator
    }

	encode_align(seqs_r, seqs_s_53);
	make_align_ptypes(seqs_r, seqs_s_53, cons_stru);

    /* ------------ calculate partition function of secondary structure --------- */
    if (constraint) {
        #pragma omp sections
        {
            #pragma omp section
            calc_secondary_pfunc(fp, seqs_r, aSR, QR, QbR, QbsR, QmR, QfR, indxR, pair_typeR, BP);
            #pragma omp section
            calc_secondary_pfunc(fp, seqs_s_53, aSS, QS, QbS, QbsS, QmS, QfS, indxS, pair_typeS, BP+len_r+1);
        }
    }
    else {
        #pragma omp sections
        {
            #pragma omp section
            calc_secondary_pfunc(fp, seqs_r, aSR, QR, QbR, QbsR, QmR, QfR, indxR, pair_typeR, NULL);
            #pragma omp section
            calc_secondary_pfunc(fp, seqs_s_53, aSS, QS, QbS, QbsS, QmS, QfS, indxS, pair_typeS, NULL);
        }
    }

    #ifndef DEBUG
    print_block_pf();
    return;
    #endif

    //adjust the order of PF of S
    for (is=1; is<=len_s/2; is++)
        for (js=is; js<=(len_s-is+1); js++) {
            double tmp;
            
            tmp = QS[indxS[js]+is];
            QS[indxS[js]+is] = QS[indxS[len_s-is+1]+(len_s-js+1)];
            QS[indxS[len_s-is+1]+(len_s-js+1)] = tmp;

            tmp = QbS[indxS[js]+is];
            QbS[indxS[js]+is] = QbS[indxS[len_s-is+1]+(len_s-js+1)];
            QbS[indxS[len_s-is+1]+(len_s-js+1)] = tmp;

            tmp = QbsS[indxS[js]+is];
            QbsS[indxS[js]+is] = QbsS[indxS[len_s-is+1]+(len_s-js+1)];
            QbsS[indxS[len_s-is+1]+(len_s-js+1)] = tmp;

            tmp = QmS[indxS[js]+is];
            QmS[indxS[js]+is] = QmS[indxS[len_s-is+1]+(len_s-js+1)];
            QmS[indxS[len_s-is+1]+(len_s-js+1)] = tmp;

            tmp = QfS[indxS[js]+is];
            QfS[indxS[js]+is] = QfS[indxS[len_s-is+1]+(len_s-js+1)];
            QfS[indxS[len_s-is+1]+(len_s-js+1)] = tmp;
        }
    //reverse the direction of parenthesis in S
    if (cons_stru) {
        len = strlen(cons_stru);
        
        for (js=len_r+2; js<=(len_r+1+(len-len_r-1)/2); js++) {
            char ch = cons_stru[js-1];
            if (ch == '(')
                ch = ')';
            else if (ch == ')')
                ch = '(';
            
            cons_stru[js-1] = cons_stru[len-js+len_r+1];
            if (cons_stru[js-1] == '(')
                cons_stru[js-1] = ')';
            else if (cons_stru[js-1] == ')')
                cons_stru[js-1] = '(';
            
            cons_stru[len-js+len_r+1] = ch;
        }

        for (ir=1; ir<(len_r+len_s+2); ir++)
            BP[ir] = 0;
        BP[len_r+1] = -10;   //separator
    }
    
    encode_align(seqs_r, seqs_s_35);
    make_align_ptypes(seqs_r, seqs_s_35, cons_stru);
    
    #ifndef      DEBUG
    print_secPfunc(seqs_r, QR, QbsR, QbR, indxR);
    print_secPfunc(seqs_s_53, QS, QbsS, QbS, indxS);
    return;
    #endif
    
    
    /* ============ begin to calculate partition function of seqence alignment ============ */
    
    //partition function of block (ir,jr;is,js)
	for (gr=1; gr<len_r; gr++)   //length of gap of R is from 1 to len_r-1
        #pragma omp parallel default (shared) private (ir,jr,irjr,gs,is,js,isjs,k,k1,k2,Tight,G,bonus_hy, \
                                                       GMR,GMS,GFR,GFS,typeR,typeS,typeI,type1,type2, \
                                                       type_I2,type_Ik,type_kR,type_kS,typeR_i,typeS_i)
        #pragma omp for schedule (static)
        for (ir=1; ir<=len_r-gr; ir++) 
            for (gs=1; gs<len_s; gs++)   //length of gap of S is from 1 to len_s-1
                for (is=1; is<=len_s-gs; is++) {
                    //printf("ID = %d\n", omp_get_thread_num());
                    
                    jr = ir+gr;
                    js = is+gs;
                    irjr = indxR[jr]+ir;
                    isjs = indxS[js]+is;

                    typeR = pair_typeR[0][irjr];
                    typeS = pair_typeS[0][isjs];
                    typeR_i = pair_typeR[0][indxR[jr-1]+ir+1];
                    typeS_i = pair_typeS[0][indxS[js-1]+is+1];

                    bonus_hy = 0;

                    typeI = pair_typeHy[0][ir][is];
                    type_I2 = pair_typeHy[0][jr][js];

                    /* partition function of Hybrid */
                    if (typeI==PAIR && type_I2==PAIR) {
                        if (constraint) {
                            if (BP[ir]==len_r+is+1 || BP[ir]==-2)
                                bonus_hy -= PF_BONUS;
                            if (BP[jr]==len_r+js+1 || BP[jr]==-2)
                                bonus_hy -= PF_BONUS;
                            if (BP[len_r+is+1]==-2)
                                bonus_hy -= PF_BONUS;
                            if (BP[len_r+js+1]==-2)
                                bonus_hy -= PF_BONUS;
                        }
                    
                        //only two interactions
                        G = 0;
                        for (k=1; k<=num_align; k++) {
                            type1 = pair_typeHy[k][ir][is];
                            type2 = pair_typeHy[k][jr][js];

                            if (type1 == 0)
                                type1 = 7;
                            if (type2 == 0)
                                type2 = 7;
                            G += calc_interior_energy(jr-ir-1, js-is-1, type1, rtype[type2],
                                                      aSR[k][ir+1], aSS[k][is+1], aSR[k][jr-1], aSS[k][js-1]);
                        }
                        Bond[0][irjr][isjs] = EXP(SIGMAE*((double) G)/num_align);
                        Bond[1][irjr][isjs] = EXP(SIGMA*((double) G)/num_align);
                        Bond[2][irjr][isjs] = EXP(SIGMA*((double) G)/num_align);
                        Bond[3][irjr][isjs] = EXP(SIGMA*((double) G)/num_align);
                        
                        for (k1=ir+1; k1<jr; k1++)
                            for (k2=is+1; k2<js; k2++) {
                                type_Ik = pair_typeHy[0][k1][k2];
                                
                                if (type_Ik == NO_PAIR)
                                    continue;

                                G = 0;
                                for (k=1; k<=num_align; k++) {
                                    type1 = pair_typeHy[k][k1][k2];
                                    type2 = pair_typeHy[k][jr][js];
                                    
                                    if (type1 == 0)
                                        type1 = 7;
                                    if (type2 == 0)
                                        type2 = 7;
                                    G += calc_interior_energy(jr-k1-1, js-k2-1, type1, rtype[type2],
                                                              aSR[k][k1+1], aSS[k][k2+1], aSR[k][jr-1], aSS[k][js-1]);
                                }
                                Bond[0][irjr][isjs] += Bond[0][indxR[k1]+ir][indxS[k2]+is]
                                    *EXP(SIGMAE*((double) G)/num_align);  //ee
                                Bond[1][irjr][isjs] += Bond[1][indxR[k1]+ir][indxS[k2]+is]
                                    *EXP(SIGMA*((double) G)/num_align);  //ke
                                Bond[2][irjr][isjs] += Bond[2][indxR[k1]+ir][indxS[k2]+is]
                                    *EXP(SIGMA*((double) G)/num_align);  //ek
                                Bond[3][irjr][isjs] += Bond[3][indxR[k1]+ir][indxS[k2]+is]
                                    *EXP(SIGMA*((double) G)/num_align);  //kk
                            }
                    }
                    else {
                        Bond[0][irjr][isjs] = 0;
                        Bond[1][irjr][isjs] = 0;
                        Bond[2][irjr][isjs] = 0;
                        Bond[3][irjr][isjs] = 0;
                    }
                    
                    if (constraint) {
                        Bond[0][irjr][isjs] *= EXP(bonus_hy);       //(Bond[0][irjr][isjs]+PF_BASE)*bonus_hy;
                        Bond[1][irjr][isjs] *= EXP(bonus_hy);       //(Bond[1][irjr][isjs]+PF_BASE)*bonus_hy;
                        Bond[2][irjr][isjs] *= EXP(bonus_hy);       //(Bond[2][irjr][isjs]+PF_BASE)*bonus_hy;
                        Bond[3][irjr][isjs] *= EXP(bonus_hy);       //(Bond[3][irjr][isjs]+PF_BASE)*bonus_hy;
                    }
                    
                    /* partition function of DT (QDT) */
                    QDT[1][irjr][isjs] = Bond[0][irjr][isjs]*EXP(SIGMAE0);
                    QDT[11][irjr][isjs] = Bond[2][irjr][isjs]*EXP(SIGMA0);
                    QDT[15][irjr][isjs] = Bond[1][irjr][isjs]*EXP(SIGMA0);
                    QDT[19][irjr][isjs] = Bond[3][irjr][isjs]*EXP(SIGMA0);
                    for (k1=ir+1; k1<jr-1; k1++) 
                        for (k2=is+1; k2<js-1; k2++) {
                            type_kR = pair_typeR[0][indxR[k1]+ir];
                            type_kS = pair_typeS[0][indxS[k2]+is];

                            GMR = GMS = 0;
                            if (type_kR == PAIR) {
                                for (k=1; k<=num_align; k++)
                                    GMR += P->MLintern[pair_typeR[k][indxR[k1]+ir]];
                            }
                            else
                                GMR = 0;
                            if (type_kS == PAIR) {
                                for (k=1; k<=num_align; k++)
                                    GMS += P->MLintern[rtype[pair_typeS[k][indxS[k2]+is]]];
                            }
                            else
                                GMS = 0;
                            GFR = GFS = BETA2;

                            //eea
                            QDT[0][irjr][isjs] += (QTu[0][indxR[k1]+ir][indxS[k2]+is]
                                                   +QTd[0][indxR[k1]+ir][indxS[k2]+is]+QTe[indxR[k1]+ir][indxS[k2]+is])
                                * (QRT[0][indxR[jr]+k1+1][indxS[js]+k2+1]+QRT[1][indxR[jr]+k1+1][indxS[js]+k2+1]);
                            //eeb
                            QDT[1][irjr][isjs] += Bond[0][indxR[k1]+ir][indxS[k2]+is] * EXP(SIGMAE0)
                                * QRT[0][indxR[jr]+k1+1][indxS[js]+k2+1];
                            //me
                            QDT[2][irjr][isjs] += (QTu[0][indxR[k1]+ir][indxS[k2]+is]*EXP(((double) GMR)/num_align)
                                                   +QTd[1][indxR[k1]+ir][indxS[k2]+is]
                                                   +QTe[indxR[k1]+ir][indxS[k2]+is]*EXP(((double) GMR)/num_align))
                                * QRT[2][indxR[jr]+k1+1][indxS[js]+k2+1];
                            //em
                            QDT[3][irjr][isjs] += (QTu[1][indxR[k1]+ir][indxS[k2]+is]
                                                   +QTd[0][indxR[k1]+ir][indxS[k2]+is]*EXP(((double) GMS)/num_align)
                                                   +QTe[indxR[k1]+ir][indxS[k2]+is]*EXP(((double) GMS)/num_align))
                                * QRT[3][indxR[jr]+k1+1][indxS[js]+k2+1];
                            //fe
                            QDT[4][irjr][isjs] += (QTu[0][indxR[k1]+ir][indxS[k2]+is]*EXP(GFR)
                                                   +QTd[2][indxR[k1]+ir][indxS[k2]+is]
                                                   +QTe[indxR[k1]+ir][indxS[k2]+is]*EXP(GFR))
                                * QRT[4][indxR[jr]+k1+1][indxS[js]+k2+1];
                            //ef
                            QDT[5][irjr][isjs] += (QTu[2][indxR[k1]+ir][indxS[k2]+is]
                                                   +QTd[0][indxR[k1]+ir][indxS[k2]+is]*EXP(GFS)
                                                   +QTe[indxR[k1]+ir][indxS[k2]+is]*EXP(GFS))
                                * QRT[5][indxR[jr]+k1+1][indxS[js]+k2+1];
                            //mm
                            QDT[6][irjr][isjs] += (QTu[1][indxR[k1]+ir][indxS[k2]+is]*EXP(((double) GMR)/num_align)
                                                   +QTd[1][indxR[k1]+ir][indxS[k2]+is]*EXP(((double) GMS)/num_align)
                                                   +QTe[indxR[k1]+ir][indxS[k2]+is]*EXP(((double) GMR)/num_align)
                                                   *EXP(((double) GMS)/num_align))
                                * QRT[6][indxR[jr]+k1+1][indxS[js]+k2+1];
                            //mf
                            QDT[7][irjr][isjs] += (QTu[2][indxR[k1]+ir][indxS[k2]+is]*EXP(((double) GMR)/num_align)
                                                   +QTd[1][indxR[k1]+ir][indxS[k2]+is]*EXP(GFS)
                                                   +QTe[indxR[k1]+ir][indxS[k2]+is]*EXP(((double) GMR)/num_align)
                                                   *EXP(GFS))
                                * QRT[7][indxR[jr]+k1+1][indxS[js]+k2+1];
                            //fm
                            QDT[8][irjr][isjs] += (QTu[1][indxR[k1]+ir][indxS[k2]+is]*EXP(GFR)
                                                   +QTd[2][indxR[k1]+ir][indxS[k2]+is]*EXP(((double) GMS)/num_align)
                                                   +QTe[indxR[k1]+ir][indxS[k2]+is]*EXP(GFR)
                                                   *EXP(((double) GMS)/num_align))
                                * QRT[8][indxR[jr]+k1+1][indxS[js]+k2+1];
                            //ff
                            QDT[9][irjr][isjs] += (QTu[2][indxR[k1]+ir][indxS[k2]+is]*EXP(GFR)
                                                   +QTd[2][indxR[k1]+ir][indxS[k2]+is]*EXP(GFS)
                                                   +QTe[indxR[k1]+ir][indxS[k2]+is]*EXP(GFR)*EXP(GFS))
                                * QRT[9][indxR[jr]+k1+1][indxS[js]+k2+1];
                            //eka
                            QDT[10][irjr][isjs] += QTu[3][indxR[k1]+ir][indxS[k2]+is]
                                * (QRT[10][indxR[jr]+k1+1][indxS[js]+k2+1]+QRT[11][indxR[jr]+k1+1][indxS[js]+k2+1]
                                   +QRT[5][indxR[jr]+k1+1][indxS[js]+k2+1])
                                + (QTu[2][indxR[k1]+ir][indxS[k2]+is]+QTd[0][indxR[k1]+ir][indxS[k2]+is]*EXP(GFS)
                                   +QTe[indxR[k1]+ir][indxS[k2]+is]*EXP(GFS))
                                * (QRT[10][indxR[jr]+k1+1][indxS[js]+k2+1]+QRT[11][indxR[jr]+k1+1][indxS[js]+k2+1]);
                            //ekb
                            QDT[11][irjr][isjs] += Bond[2][indxR[k1]+ir][indxS[k2]+is] * EXP(SIGMA0)
                                * (QRT[10][indxR[jr]+k1+1][indxS[js]+k2+1]+QRT[5][indxR[jr]+k1+1][indxS[js]+k2+1]);
                            //mk
                            QDT[12][irjr][isjs] += QTu[3][indxR[k1]+ir][indxS[k2]+is]*EXP(((double) GMR)/num_align)
                                * (QRT[12][indxR[jr]+k1+1][indxS[js]+k2+1]+QRT[7][indxR[jr]+k1+1][indxS[js]+k2+1])
                                + (QTu[2][indxR[k1]+ir][indxS[k2]+is]*EXP(((double) GMR)/num_align)
                                   +QTd[1][indxR[k1]+ir][indxS[k2]+is]*EXP(GFS)
                                   +QTe[indxR[k1]+ir][indxS[k2]+is]*EXP(((double) GMR)/num_align)*EXP(GFS))
                                * QRT[12][indxR[jr]+k1+1][indxS[js]+k2+1];
                            //fk
                            QDT[13][irjr][isjs] += QTu[3][indxR[k1]+ir][indxS[k2]+is]*EXP(GFR)
                                * (QRT[13][indxR[jr]+k1+1][indxS[js]+k2+1]+QRT[9][indxR[jr]+k1+1][indxS[js]+k2+1])
                                + (QTu[2][indxR[k1]+ir][indxS[k2]+is]*EXP(GFR)
                                   +QTd[2][indxR[k1]+ir][indxS[k2]+is]*EXP(GFS)
                                   +QTe[indxR[k1]+ir][indxS[k2]+is]*EXP(GFR)*EXP(GFS))
                                * QRT[13][indxR[jr]+k1+1][indxS[js]+k2+1];
                            //kea
                            QDT[14][irjr][isjs] += QTd[3][indxR[k1]+ir][indxS[k2]+is]
                                * (QRT[14][indxR[jr]+k1+1][indxS[js]+k2+1]+QRT[15][indxR[jr]+k1+1][indxS[js]+k2+1]
                                   +QRT[4][indxR[jr]+k1+1][indxS[js]+k2+1])
                                + (QTu[0][indxR[k1]+ir][indxS[k2]+is]*EXP(GFR)+QTd[2][indxR[k1]+ir][indxS[k2]+is]
                                   +QTe[indxR[k1]+ir][indxS[k2]+is]*EXP(GFR))
                                * (QRT[14][indxR[jr]+k1+1][indxS[js]+k2+1]+QRT[15][indxR[jr]+k1+1][indxS[js]+k2+1]);
                            //keb
                            QDT[15][irjr][isjs] += Bond[1][indxR[k1]+ir][indxS[k2]+is]*EXP(SIGMA0)
                                * (QRT[14][indxR[jr]+k1+1][indxS[js]+k2+1]
                                   +QRT[4][indxR[jr]+k1+1][indxS[js]+k2+1]);
                            //km
                            QDT[16][irjr][isjs] += QTd[3][indxR[k1]+ir][indxS[k2]+is] * EXP(((double) GMS)/num_align)
                                * (QRT[16][indxR[jr]+k1+1][indxS[js]+k2+1]+QRT[8][indxR[jr]+k1+1][indxS[js]+k2+1])
                                + (QTu[1][indxR[k1]+ir][indxS[k2]+is]*EXP(GFR)
                                   +QTd[2][indxR[k1]+ir][indxS[k2]+is]*EXP(((double) GMS)/num_align)
                                   +QTe[indxR[k1]+ir][indxS[k2]+is]*EXP(GFR)*EXP(((double) GMS)/num_align))
                                * QRT[16][indxR[jr]+k1+1][indxS[js]+k2+1];
                            //kf
                            QDT[17][irjr][isjs] += QTd[3][indxR[k1]+ir][indxS[k2]+is] * EXP(GFS)
                                * (QRT[17][indxR[jr]+k1+1][indxS[js]+k2+1]+QRT[9][indxR[jr]+k1+1][indxS[js]+k2+1])
                                + (QTu[2][indxR[k1]+ir][indxS[k2]+is]*EXP(GFR)
                                   +QTd[2][indxR[k1]+ir][indxS[k2]+is]*EXP(GFS)
                                   +QTe[indxR[k1]+ir][indxS[k2]+is]*EXP(GFR)*EXP(GFS))
                                * QRT[17][indxR[jr]+k1+1][indxS[js]+k2+1];
                            //kka
                            QDT[18][irjr][isjs] += QTu[3][indxR[k1]+ir][indxS[k2]+is] * EXP(GFR)
                                * (QRT[17][indxR[jr]+k1+1][indxS[js]+k2+1]+QRT[18][indxR[jr]+k1+1][indxS[js]+k2+1]
                                   +QRT[19][indxR[jr]+k1+1][indxS[js]+k2+1])
                                + QTd[3][indxR[k1]+ir][indxS[k2]+is]*EXP(GFS)
                                * (QRT[13][indxR[jr]+k1+1][indxS[js]+k2+1]+QRT[18][indxR[jr]+k1+1][indxS[js]+k2+1]
                                   +QRT[19][indxR[jr]+k1+1][indxS[js]+k2+1])
                                + (QTu[2][indxR[k1]+ir][indxS[k2]+is]*EXP(GFR)
                                   +QTd[2][indxR[k1]+ir][indxS[k2]+is]*EXP(GFS)
                                   +QTe[indxR[k1]+ir][indxS[k2]+is]*EXP(GFR)*EXP(GFS))
                                * (QRT[18][indxR[jr]+k1+1][indxS[js]+k2+1]+QRT[19][indxR[jr]+k1+1][indxS[js]+k2+1]);
                            //kkb
                            QDT[19][irjr][isjs] += Bond[3][indxR[k1]+ir][indxS[k2]+is]*EXP(SIGMA0)
                                *(QRT[18][indxR[jr]+k1+1][indxS[js]+k2+1]
                                  +QRT[13][indxR[jr]+k1+1][indxS[js]+k2+1]
                                  +QRT[17][indxR[jr]+k1+1][indxS[js]+k2+1]
                                  +QRT[9][indxR[jr]+k1+1][indxS[js]+k2+1]);
                        }

                    /* partition function os Tsu, Tsd and Tse */
                    // PF of Tsu
                    if (typeR==PAIR && typeR_i==NO_PAIR) {
                        QTsu[0][irjr][isjs] = calc_Tu_pfunc(QTu[0], QTsu[0], QDT[2], QDT[14], QDT[15],
                                                            ir, jr, is, js, 2);         //e
                        QTsu[1][irjr][isjs] = calc_Tu_pfunc(QTu[1], QTsu[1], QDT[6], QDT[16], NULL,
                                                            ir, jr, is, js, 2);         //m  
                        QTsu[2][irjr][isjs] = calc_Tu_pfunc(QTu[2], QTsu[2], QDT[7], QDT[17], NULL,
                                                            ir, jr, is, js, 2);         //f
                        QTsu[3][irjr][isjs] = calc_Tu_pfunc(QTu[3], QTsu[3], QDT[12], QDT[18], QDT[19],
                                                            ir, jr, is, js, 2);         //k
                    }
                    else {
                        QTsu[0][irjr][isjs] = 0.0;
                        QTsu[1][irjr][isjs] = 0.0; 
                        QTsu[2][irjr][isjs] = 0.0;
                        QTsu[3][irjr][isjs] = 0.0;
                    }
                    // PF of Tsd
                    if (typeS==PAIR && typeS_i==NO_PAIR) {
                        QTsd[0][irjr][isjs] = calc_Td_pfunc(QTd[0], QTsd[0], QDT[3], QDT[10], QDT[11],
                                                            ir, jr, is, js, 2);
                        QTsd[1][irjr][isjs] = calc_Td_pfunc(QTd[1], QTsd[1], QDT[6], QDT[12], NULL,
                                                            ir, jr, is, js, 2);
                        QTsd[2][irjr][isjs] = calc_Td_pfunc(QTd[2], QTsd[2], QDT[8], QDT[13], NULL,
                                                            ir, jr, is, js, 2);
                        QTsd[3][irjr][isjs] = calc_Td_pfunc(QTd[3], QTsd[3], QDT[16], QDT[18], QDT[19],
                                                            ir, jr, is, js, 2);
                    }
                    else {
                        QTsd[0][irjr][isjs] = 0.0;
                        QTsd[1][irjr][isjs] = 0.0; 
                        QTsd[2][irjr][isjs] = 0.0;
                        QTsd[3][irjr][isjs] = 0.0;
                    }
                    // PF of Tse
                    if (typeR==PAIR && typeS==PAIR && typeR_i==NO_PAIR && typeS_i==PAIR) {
                        QTse[irjr][isjs] = calc_Tu_pfunc(QTe, QTse, QTd[1], QTd[3], NULL, ir, jr, is, js, 2);
                    }
                    else 
                        QTse[irjr][isjs]=0.0;

                    /* partition function of Tu, Td and Te */
                    //PF of Tu
                    if (typeR==PAIR && typeR_i==PAIR) { 
                        QTu[0][irjr][isjs] = calc_Tu_pfunc(QTu[0], QTsu[0], QDT[2], QDT[14], QDT[15],
                                                           ir, jr, is, js, 1);          //e
                        QTu[1][irjr][isjs] = calc_Tu_pfunc(QTu[1], QTsu[1], QDT[6], QDT[16], NULL,
                                                           ir, jr, is, js, 1);          //m  
                        QTu[2][irjr][isjs] = calc_Tu_pfunc(QTu[2], QTsu[2], QDT[7], QDT[17], NULL,
                                                           ir, jr, is, js, 1);          //f
                        QTu[3][irjr][isjs] = calc_Tu_pfunc(QTu[3], QTsu[3], QDT[12], QDT[18], QDT[19],
                                                           ir, jr, is, js, 1);          //k
                    }
                    else {
                        QTu[0][irjr][isjs] = 0.0;
                        QTu[1][irjr][isjs] = 0.0; 
                        QTu[2][irjr][isjs] = 0.0;
                        QTu[3][irjr][isjs] = 0.0;
                    }
                    //PF of Td
                    if (typeS==PAIR && typeS_i==PAIR) { 
                        QTd[0][irjr][isjs] = calc_Td_pfunc(QTd[0], QTsd[0], QDT[3], QDT[10], QDT[11],
                                                           ir, jr, is, js, 1);
                        QTd[1][irjr][isjs] = calc_Td_pfunc(QTd[1], QTsd[1], QDT[6], QDT[12], NULL,
                                                           ir, jr, is, js, 1);
                        QTd[2][irjr][isjs] = calc_Td_pfunc(QTd[2], QTsd[2], QDT[8], QDT[13], NULL,
                                                           ir, jr, is, js, 1);
                        QTd[3][irjr][isjs] = calc_Td_pfunc(QTd[3], QTsd[3], QDT[16], QDT[18], QDT[19],
                                                           ir, jr, is, js, 1);
                    }
                    else {
                        QTd[0][irjr][isjs] = 0.0;
                        QTd[1][irjr][isjs] = 0.0; 
                        QTd[2][irjr][isjs] = 0.0;
                        QTd[3][irjr][isjs] = 0.0;
                    }
                    //PF of Te
                    if (typeR==PAIR && typeS==PAIR && typeR_i==PAIR && typeS_i==PAIR) {
                        QTe[irjr][isjs] = calc_Tu_pfunc(QTe, QTse, QTd[1], QTd[3], NULL, ir, jr, is, js, 1);
                    }
                    else 
                        QTe[irjr][isjs]=0.0;

                    /* partition function of RT */
                    //eea
                    Tight = QTu[0][irjr][isjs] + QTd[0][irjr][isjs] + QTe[irjr][isjs];
                    QRT[0][irjr][isjs] = Tight + QDT[0][irjr][isjs];
                    for (k1=ir; k1<jr-1; k1++)
                        for (k2=is; k2<js-1; k2++) {
                            Tight = QTu[0][indxR[jr]+k1+1][indxS[js]+k2+1]+QTd[0][indxR[jr]+k1+1][indxS[js]+k2+1]
                                + QTe[indxR[jr]+k1+1][indxS[js]+k2+1];
                            QRT[0][irjr][isjs] += QR[indxR[k1]+ir]*QS[indxS[k2]+is]
                                * (Tight+QDT[0][indxR[jr]+k1+1][indxS[js]+k2+1]);
                            QRT[0][irjr][isjs] += ((QR[indxR[k1]+ir]-1)*QS[indxS[k2]+is]+1*(QS[indxS[k2]+is]-1))
                                * QDT[1][indxR[jr]+k1+1][indxS[js]+k2+1];
                        }
                    for (k1=ir; k1<jr-1; k1++) {        //secondary structure on S is empty
                        Tight = QTu[0][indxR[jr]+k1+1][isjs] + QTd[0][indxR[jr]+k1+1][isjs]
                            + QTe[indxR[jr]+k1+1][isjs];
                        QRT[0][irjr][isjs] += QR[indxR[k1]+ir] * (Tight+QDT[0][indxR[jr]+k1+1][isjs]);
                        QRT[0][irjr][isjs] += (QR[indxR[k1]+ir]-1) * QDT[1][indxR[jr]+k1+1][isjs];
                    }
                    for (k2=is; k2<js-1; k2++) {        //secondary structure on R is empty
                        Tight = QTu[0][irjr][indxS[js]+k2+1]+QTd[0][irjr][indxS[js]+k2+1]
                            + QTe[irjr][indxS[js]+k2+1];
                        QRT[0][irjr][isjs] += QS[indxS[k2]+is] * (Tight+QDT[0][irjr][indxS[js]+k2+1]);
                        QRT[0][irjr][isjs] += (QS[indxS[k2]+is]-1) * QDT[1][irjr][indxS[js]+k2+1];
                    }
                    //eeb
                    for (k1=ir; k1<jr; k1++)
                        for (k2=is; k2<js; k2++)
                            QRT[1][irjr][isjs] += QDT[1][indxR[jr]+k1][indxS[js]+k2];

                    //me
                    G = 0;
                    for (k=1; k<=num_align; k++)
                        G += P->MLintern[pair_typeR[k][irjr]];
                    Tight = QTu[0][irjr][isjs]*EXP(((double) G)/num_align)+QTd[1][irjr][isjs]
                        + QTe[irjr][isjs]*EXP(((double) G)/num_align);
                    QRT[2][irjr][isjs] = Tight + QDT[2][irjr][isjs];
                    for (k1=ir; k1<jr-1; k1++)
                        for (k2=is; k2<js-1; k2++) {
                            GMR = 0;
                            for (k=1; k<=num_align; k++)
                                GMR += P->MLintern[pair_typeR[k][indxR[jr]+k1+1]];
                            
                            Tight = QTu[0][indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(((double) GMR)/num_align)
                                + QTd[1][indxR[jr]+k1+1][indxS[js]+k2+1]
                                + QTe[indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(((double) GMR)/num_align);
                            QRT[2][irjr][isjs] += (QmR[indxR[k1]+ir]+EXP((k1-ir+1)*P->MLbase))
                                * QS[indxS[k2]+is] * (Tight+QDT[2][indxR[jr]+k1+1][indxS[js]+k2+1]);
                        }
                    for (k1=ir; k1<jr-1; k1++) {
                        GMR = 0;
                        for (k=1; k<=num_align; k++)
                            GMR += P->MLintern[pair_typeR[k][indxR[jr]+k1+1]];
                        
                        Tight = QTu[0][indxR[jr]+k1+1][isjs]*EXP(((double) GMR)/num_align)
                            + QTd[1][indxR[jr]+k1+1][isjs] + QTe[indxR[jr]+k1+1][isjs]*EXP(((double) GMR)/num_align);
                        QRT[2][irjr][isjs] += (QmR[indxR[k1]+ir]+EXP((k1-ir+1)*P->MLbase))
                            * (Tight+QDT[2][indxR[jr]+k1+1][isjs]);
                    }
                    for (k2=is; k2<js-1; k2++) {
                        GMR = 0;
                        for (k=1; k<=num_align; k++)
                            GMR += P->MLintern[pair_typeR[k][irjr]];
                        
                        Tight = QTu[0][irjr][indxS[js]+k2+1]*EXP(((double) GMR)/num_align)
                            + QTd[1][irjr][indxS[js]+k2+1] + QTe[irjr][indxS[js]+k2+1]*EXP(((double) GMR)/num_align);
                        QRT[2][irjr][isjs] += QS[indxS[k2]+is] * (Tight+QDT[2][irjr][indxS[js]+k2+1]);
                    }
                    //em
                    G = 0;
                    for (k=1; k<=num_align; k++)
                        G += P->MLintern[rtype[pair_typeS[k][isjs]]];
                    Tight = QTu[1][irjr][isjs] + QTd[0][irjr][isjs]*EXP(((double) G)/num_align)
                        + QTe[irjr][isjs]*EXP(((double) G)/num_align);
                    QRT[3][irjr][isjs] = Tight + QDT[3][irjr][isjs];
                    for (k1=ir; k1<jr-1; k1++)
                        for (k2=is; k2<js-1; k2++) {
                            GMS = 0;
                            for (k=1; k<=num_align; k++)
                                GMS += P->MLintern[rtype[pair_typeS[k][indxS[js]+k2+1]]];
                            
                            Tight = QTu[1][indxR[jr]+k1+1][indxS[js]+k2+1]
                                + QTd[0][indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(((double) GMS)/num_align)
                                + QTe[indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(((double) GMS)/num_align);
                            QRT[3][irjr][isjs] += QR[indxR[k1]+ir]
                                * (QmS[indxS[k2]+is]+EXP((k2-is+1)*P->MLbase))
                                * (Tight+QDT[3][indxR[jr]+k1+1][indxS[js]+k2+1]);
                        }
                    for (k1=ir; k1<jr-1; k1++) {
                        GMS = 0;
                        for (k=1; k<=num_align; k++)
                            GMS += P->MLintern[rtype[pair_typeS[k][isjs]]];
                        
                        Tight = QTu[1][indxR[jr]+k1+1][isjs]
                            + QTd[0][indxR[jr]+k1+1][isjs]*EXP(((double) GMS)/num_align)
                            + QTe[indxR[jr]+k1+1][isjs]*EXP(((double) GMS)/num_align);
                        QRT[3][irjr][isjs] += QR[indxR[k1]+ir] * (Tight+QDT[3][indxR[jr]+k1+1][isjs]);
                    }
                    for (k2=is; k2<js-1; k2++) {
                        GMS = 0;
                        for (k=1; k<=num_align; k++)
                            GMS += P->MLintern[rtype[pair_typeS[k][indxS[js]+k2+1]]];
                        
                        Tight = QTu[1][irjr][indxS[js]+k2+1]
                            + QTd[0][irjr][indxS[js]+k2+1]*EXP(((double) GMS)/num_align)
                            + QTe[irjr][indxS[js]+k2+1]*EXP(((double) GMS)/num_align);
                        QRT[3][irjr][isjs] += (QmS[indxS[k2]+is]+EXP((k2-is+1)*P->MLbase))
                            * (Tight+QDT[3][irjr][indxS[js]+k2+1]);
                    }
                    //fe
                    Tight = QTu[0][irjr][isjs]*EXP(BETA2) + QTd[2][irjr][isjs]
                        + QTe[irjr][isjs]*EXP(BETA2);
                    QRT[4][irjr][isjs] = Tight + QDT[4][irjr][isjs];
                    for (k1=ir; k1<jr-1; k1++)
                        for (k2=is; k2<js-1; k2++) {
                            GFR = BETA2;
                            
                            Tight = QTu[0][indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(GFR)
                                + QTd[2][indxR[jr]+k1+1][indxS[js]+k2+1]
                                + QTe[indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(GFR);
                            QRT[4][irjr][isjs] += (QfR[indxR[k1]+ir]+EXP((k1-ir+1)*BETA3))
                                * QS[indxS[k2]+is] * (Tight+QDT[4][indxR[jr]+k1+1][indxS[js]+k2+1]);
                        }
                    for (k1=ir; k1<jr-1; k1++) {
                        GFR = BETA2;
                        
                        Tight = QTu[0][indxR[jr]+k1+1][isjs]*EXP(GFR)
                            + QTd[2][indxR[jr]+k1+1][isjs] + QTe[indxR[jr]+k1+1][isjs]*EXP(GFR);
                        QRT[4][irjr][isjs] += (QfR[indxR[k1]+ir]+EXP((k1-ir+1)*BETA3))
                            * (Tight+QDT[4][indxR[jr]+k1+1][isjs]);
                    }
                    for (k2=is; k2<js-1; k2++) {
                        GFR = BETA2;
                        
                        Tight = QTu[0][irjr][indxS[js]+k2+1]*EXP(GFR)
                            + QTd[2][irjr][indxS[js]+k2+1] + QTe[irjr][indxS[js]+k2+1]*EXP(GFR);
                        QRT[4][irjr][isjs] += QS[indxS[k2]+is] * (Tight+QDT[4][irjr][indxS[js]+k2+1]);
                    }
                    //ef
                    Tight = QTu[2][irjr][isjs] + QTd[0][irjr][isjs]*EXP(BETA2)
                        + QTe[irjr][isjs]*EXP(BETA2);
                    QRT[5][irjr][isjs] = Tight + QDT[5][irjr][isjs];
                    for (k1=ir; k1<jr-1; k1++)
                        for (k2=is; k2<js-1; k2++) {
                            GFS = BETA2;
                            
                            Tight = QTu[2][indxR[jr]+k1+1][indxS[js]+k2+1]
                                + QTd[0][indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(GFS)
                                + QTe[indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(GFS);
                            QRT[5][irjr][isjs] += QR[indxR[k1]+ir]
                                * (QfS[indxS[k2]+is]+EXP((k2-is+1)*BETA3))
                                * (Tight+QDT[5][indxR[jr]+k1+1][indxS[js]+k2+1]);
                        }
                    for (k1=ir; k1<jr-1; k1++) {
                        GFS = BETA2;
                        
                        Tight = QTu[2][indxR[jr]+k1+1][isjs]
                            + QTd[0][indxR[jr]+k1+1][isjs]*EXP(GFS) + QTe[indxR[jr]+k1+1][isjs]*EXP(GFS);
                        QRT[5][irjr][isjs] += QR[indxR[k1]+ir] * (Tight+QDT[5][indxR[jr]+k1+1][isjs]);
                    }
                    for (k2=is; k2<js-1; k2++) {
                        GFS = BETA2;
                        
                        Tight = QTu[2][irjr][indxS[js]+k2+1]
                            + QTd[0][irjr][indxS[js]+k2+1]*EXP(GFS) + QTe[irjr][indxS[js]+k2+1]*EXP(GFS);
                        QRT[5][irjr][isjs] += (QfS[indxS[k2]+is]+EXP((k2-is+1)*BETA3))
                            * (Tight+QDT[5][irjr][indxS[js]+k2+1]);
                    }
                    //mm
                    GMR = GMS = 0;
                    for (k=1; k<=num_align; k++) {
                        GMR += P->MLintern[pair_typeR[k][irjr]];
                        GMS += P->MLintern[rtype[pair_typeS[k][isjs]]];
                    }
                    Tight = QTu[1][irjr][isjs]*EXP(((double) GMR)/num_align)
                        + QTd[1][irjr][isjs]*EXP(((double) GMS)/num_align)
                        + QTe[irjr][isjs]*EXP(((double) GMR)/num_align)*EXP(((double) GMS)/num_align);
                    QRT[6][irjr][isjs] = Tight + QDT[6][irjr][isjs];
                    for (k1=ir; k1<jr-1; k1++)
                        for (k2=is; k2<js-1; k2++) {
                            GMR = GMS = 0;
                            for (k=1; k<=num_align; k++) {
                                GMR += P->MLintern[pair_typeR[k][indxR[jr]+k1+1]];
                                GMS += P->MLintern[rtype[pair_typeS[k][indxS[js]+k2+1]]];
                            }
                            
                            Tight = QTu[1][indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(((double) GMR)/num_align)
                                +QTd[1][indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(((double) GMS)/num_align)
                                +QTe[indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(((double) GMR)/num_align)
                                *EXP(((double) GMS)/num_align);
                            QRT[6][irjr][isjs] += (QmR[indxR[k1]+ir]+EXP((k1-ir+1)*P->MLbase))
                                * (QmS[indxS[k2]+is]+EXP((k2-is+1)*P->MLbase))
                                * (Tight+QDT[6][indxR[jr]+k1+1][indxS[js]+k2+1]);
                        }
                    for (k1=ir; k1<jr-1; k1++) {
                        GMR = GMS = 0;
                        for (k=1; k<=num_align; k++) {
                            GMR += P->MLintern[pair_typeR[k][indxR[jr]+k1+1]];
                            GMS += P->MLintern[rtype[pair_typeS[k][isjs]]];
                        }
                        
                        Tight = QTu[1][indxR[jr]+k1+1][isjs]*EXP(((double) GMR)/num_align)
                            + QTd[1][indxR[jr]+k1+1][isjs]*EXP(((double) GMS)/num_align)
                            + QTe[indxR[jr]+k1+1][isjs]*EXP(((double) GMR)/num_align)*EXP(((double) GMS)/num_align);
                        QRT[6][irjr][isjs] += (QmR[indxR[k1]+ir]+EXP((k1-ir+1)*P->MLbase))
                            * (Tight+QDT[6][indxR[jr]+k1+1][isjs]);
                    }
                    for (k2=is; k2<js-1; k2++) {
                        GMR = GMS = 0;
                        for (k=1; k<=num_align; k++) {
                            GMR += P->MLintern[pair_typeR[k][irjr]];
                            GMS += P->MLintern[rtype[pair_typeS[k][indxS[js]+k2+1]]];
                        }
                        
                        Tight = QTu[1][irjr][indxS[js]+k2+1]*EXP(((double) GMR)/num_align)
                            + QTd[1][irjr][indxS[js]+k2+1]*EXP(((double) GMS)/num_align)
                            + QTe[irjr][indxS[js]+k2+1]*EXP(((double) GMR)/num_align)*EXP(((double) GMS)/num_align);
                        QRT[6][irjr][isjs] += (QmS[indxS[k2]+is]+EXP((k2-is+1)*P->MLbase))
                            * (Tight+QDT[6][irjr][indxS[js]+k2+1]);
                    }
                    //mf
                    G = 0;
                    for (k=1; k<=num_align; k++)
                        G += P->MLintern[pair_typeR[k][irjr]];
                    
                    Tight = QTu[2][irjr][isjs]*EXP(((double) G)/num_align) + QTd[1][irjr][isjs]*EXP(BETA2)
                        + QTe[irjr][isjs]*EXP(((double) G)/num_align)*EXP(BETA2);
                    QRT[7][irjr][isjs] = Tight + QDT[7][irjr][isjs];
                    for (k1=ir; k1<jr-1; k1++)
                        for (k2=is; k2<js-1; k2++) {
                            GMR = 0;
                            for (k=1; k<=num_align; k++)
                                GMR += P->MLintern[pair_typeR[k][indxR[jr]+k1+1]];
                            GFS = BETA2;
                            
                            Tight = QTu[2][indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(((double) GMR)/num_align)
                                + QTd[1][indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(GFS)
                                + QTe[indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(((double) GMR)/num_align)*EXP(GFS);
                            QRT[7][irjr][isjs] += (QmR[indxR[k1]+ir]+EXP((k1-ir+1)*P->MLbase))
                                * (QfS[indxS[k2]+is]+EXP((k2-is+1)*BETA3))
                                * (Tight+QDT[7][indxR[jr]+k1+1][indxS[js]+k2+1]);
                        }
                    for (k1=ir; k1<jr-1; k1++) {
                        GMR = 0;
                        for (k=1; k<=num_align; k++)
                            GMR += P->MLintern[pair_typeR[k][indxR[jr]+k1+1]];
                        GFS = BETA2;
                        
                        Tight = QTu[2][indxR[jr]+k1+1][isjs]*EXP(((double) GMR)/num_align)
                            + QTd[1][indxR[jr]+k1+1][isjs]*EXP(GFS)
                            + QTe[indxR[jr]+k1+1][isjs]*EXP(((double) GMR)/num_align)*EXP(GFS);
                        QRT[7][irjr][isjs] += (QmR[indxR[k1]+ir]+EXP((k1-ir+1)*P->MLbase))
                            * (Tight+QDT[7][indxR[jr]+k1+1][isjs]);
                    }
                    for (k2=is; k2<js-1; k2++) {
                        GMR = 0;
                        for (k=1; k<=num_align; k++)
                            GMR += P->MLintern[pair_typeR[k][irjr]];
                        GFS = BETA2;
                        
                        Tight = QTu[2][irjr][indxS[js]+k2+1]*EXP(((double) GMR)/num_align)
                            + QTd[1][irjr][indxS[js]+k2+1]*EXP(GFS)
                            + QTe[irjr][indxS[js]+k2+1]*EXP(((double) GMR)/num_align)*EXP(GFS);
                        QRT[7][irjr][isjs] += (QfS[indxS[k2]+is]+EXP((k2-is+1)*BETA3))
                            * (Tight+QDT[7][irjr][indxS[js]+k2+1]);
                    }
                    //fm
                    G = 0;
                    for (k=1; k<=num_align; k++)
                        G += P->MLintern[rtype[pair_typeS[k][isjs]]];
                    
                    Tight = QTu[1][irjr][isjs]*EXP(BETA2) + QTd[2][irjr][isjs]*EXP(((double) G)/num_align)
                        + QTe[irjr][isjs]*EXP(BETA2)*EXP(((double) G)/num_align);
                    QRT[8][irjr][isjs] = Tight + QDT[8][irjr][isjs];
                    for (k1=ir; k1<jr-1; k1++)
                        for (k2=is; k2<js-1; k2++) {
                            GFR = BETA2;
                            GMS = 0;
                            for (k=1; k<=num_align; k++)
                                GMS += P->MLintern[rtype[pair_typeS[k][indxS[js]+k2+1]]];
                            
                            Tight = QTu[1][indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(GFR)
                                + QTd[2][indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(((double) GMS)/num_align)
                                + QTe[indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(GFR)*EXP(((double) GMS)/num_align);
                            QRT[8][irjr][isjs] += (QfR[indxR[k1]+ir]+EXP((k1-ir+1)*BETA3))
                                * (QmS[indxS[k2]+is]+EXP((k2-is+1)*P->MLbase))
                                * (Tight+QDT[8][indxR[jr]+k1+1][indxS[js]+k2+1]);
                        }
                    for (k1=ir; k1<jr-1; k1++) {
                        GFR = BETA2;
                        GMS = 0;
                        for (k=1; k<=num_align; k++)
                            GMS += P->MLintern[rtype[pair_typeS[k][isjs]]];
                        
                        Tight = QTu[1][indxR[jr]+k1+1][isjs]*EXP(GFR)
                            + QTd[2][indxR[jr]+k1+1][isjs]*EXP(((double) GMS)/num_align)
                            + QTe[indxR[jr]+k1+1][isjs]*EXP(GFR)*EXP(((double) GMS)/num_align);
                        QRT[8][irjr][isjs] += (QfR[indxR[k1]+ir]+EXP((k1-ir+1)*BETA3))
                            * (Tight+QDT[8][indxR[jr]+k1+1][isjs]);
                    }
                    for (k2=is; k2<js-1; k2++) {
                        GMR = BETA2;
                        GMS = 0;
                        for (k=1; k<=num_align; k++)
                            GMS += P->MLintern[rtype[pair_typeS[k][indxS[js]+k2+1]]];
                        
                        Tight = QTu[1][irjr][indxS[js]+k2+1]*EXP(GFR)
                            + QTd[2][irjr][indxS[js]+k2+1]*EXP(((double) GMS)/num_align)
                            + QTe[irjr][indxS[js]+k2+1]*EXP(GFR)*EXP(((double) GMS)/num_align);
                        QRT[8][irjr][isjs] += (QmS[indxS[k2]+is]+EXP((k2-is+1)*P->MLbase))
                            * (Tight+QDT[8][irjr][indxS[js]+k2+1]);
                    }
                    //ff
                    Tight = QTu[2][irjr][isjs]*EXP(BETA2) + QTd[2][irjr][isjs]*EXP(BETA2)
                        + QTe[irjr][isjs]*EXP(BETA2)*EXP(BETA2);
                    QRT[9][irjr][isjs] = Tight + QDT[9][irjr][isjs];
                    for (k1=ir; k1<jr-1; k1++)
                        for (k2=is; k2<js-1; k2++) {
                            GFR = GFS = BETA2;
                            
                            Tight = QTu[2][indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(GFR)
                                + QTd[2][indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(GFS)
                                + QTe[indxR[jr]+k1+1][indxS[js]+k2+1]*EXP(GFR)*EXP(GFS);
                            QRT[9][irjr][isjs] += (QfR[indxR[k1]+ir]+EXP((k1-ir+1)*BETA3))
                                * (QfS[indxS[k2]+is]+EXP((k2-is+1)*BETA3))
                                * (Tight+QDT[9][indxR[jr]+k1+1][indxS[js]+k2+1]);
                        }
                    for (k1=ir; k1<jr-1; k1++) {
                        GFR = GFS = BETA2;
                        
                        Tight = QTu[2][indxR[jr]+k1+1][isjs]*EXP(GFR)
                            + QTd[2][indxR[jr]+k1+1][isjs]*EXP(GFS) + QTe[indxR[jr]+k1+1][isjs]*EXP(GFR)*EXP(GFS);
                        QRT[9][irjr][isjs] += (QfR[indxR[k1]+ir]+EXP((k1-ir+1)*BETA3))
                            * (Tight+QDT[9][indxR[jr]+k1+1][isjs]);
                    }
                    for (k2=is; k2<js-1; k2++) {
                        GMR = GFS = BETA2;
                        
                        Tight = QTu[2][irjr][indxS[js]+k2+1]*EXP(GFR)
                            + QTd[2][irjr][indxS[js]+k2+1]*EXP(GFS) + QTe[irjr][indxS[js]+k2+1]*EXP(GFR)*EXP(GFS);
                        QRT[9][irjr][isjs] += (QfS[indxS[k2]+is]+EXP((k2-is+1)*BETA3))
                            *(Tight+QDT[9][irjr][indxS[js]+k2+1]);
                    }
                    //eka
                    Tight = QTu[3][irjr][isjs];
                    QRT[10][irjr][isjs] = Tight + QDT[10][irjr][isjs];
                    for (k1=ir; k1<jr-1; k1++)
                        for (k2=is; k2<js-1; k2++) {
                            Tight = QTu[3][indxR[jr]+k1+1][indxS[js]+k2+1];
                            QRT[10][irjr][isjs] += QR[indxR[k1]+ir]
                                * (QfS[indxS[k2]+is]+EXP((k2-is+1)*BETA3))
                                * (Tight+QDT[10][indxR[jr]+k1+1][indxS[js]+k2+1]);
                            QRT[10][irjr][isjs] += (QR[indxR[k1]+ir]*QfS[indxS[k2]+is]
                                                    +(QR[indxR[k1]+ir]-1)*EXP((k2-is+1)*BETA3))
                                * QDT[11][indxR[jr]+k1+1][indxS[js]+k2+1];
                        }
                    for (k1=ir; k1<jr-1; k1++) {
                        Tight = QTu[3][indxR[jr]+k1+1][isjs];
                        QRT[10][irjr][isjs] += QR[indxR[k1]+ir] * (Tight+QDT[10][indxR[jr]+k1+1][isjs]);
                        QRT[10][irjr][isjs] += (QR[indxR[k1]+ir]-1) * QDT[11][indxR[jr]+k1+1][isjs];
                    }
                    for (k2=is; k2<js-1; k2++) {
                        Tight = QTu[3][irjr][indxS[js]+k2+1];
                        QRT[10][irjr][isjs] += (QfS[indxS[k2]+is]+EXP((k2-is+1)*BETA3))
                            * (Tight+QDT[10][irjr][indxS[js]+k2+1]);
                        QRT[10][irjr][isjs] += QfS[indxS[k2]+is] * QDT[11][irjr][indxS[js]+k2+1];
                    }
                    //ekb
                    for (k1=ir; k1<jr; k1++)
                        for (k2=is; k2<js; k2++)
                            QRT[11][irjr][isjs] += QDT[11][indxR[jr]+k1][indxS[js]+k2]
                                * EXP((k2-is)*BETA3);
                    //mk
                    G = 0;
                    for (k=1; k<=num_align; k++)
                        G += P->MLintern[pair_typeR[k][irjr]];
                    
                    Tight = QTu[3][irjr][isjs]*EXP(((double) G)/num_align);
                    QRT[12][irjr][isjs] = Tight + QDT[12][irjr][isjs];
                    for (k1=ir; k1<jr-1; k1++)
                        for (k2=is; k2<js-1; k2++) {
                            GMR = 0;
                            for (k=1; k<=num_align; k++)
                                GMR += P->MLintern[pair_typeR[k][indxR[jr]+k1+1]];
                            
                            Tight = QTu[3][indxR[jr]+k1+1][indxS[js]+k2+1] * EXP(((double) GMR)/num_align);
                            QRT[12][irjr][isjs] += (QmR[indxR[k1]+ir]+EXP((k1-ir+1)*P->MLbase))
                                * (QfS[indxS[k2]+is]+EXP((k2-is+1)*BETA3))
                                * (Tight+QDT[12][indxR[jr]+k1+1][indxS[js]+k2+1]);
                        }
                    for (k1=ir; k1<jr-1; k1++) {
                        GMR = 0;
                        for (k=1; k<=num_align; k++)
                            GMR += P->MLintern[pair_typeR[k][indxR[jr]+k1+1]];
                        
                        Tight = QTu[3][indxR[jr]+k1+1][isjs] * EXP(((double) GMR)/num_align);
                        QRT[12][irjr][isjs] += (QmR[indxR[k1]+ir]+EXP((k1-ir+1)*P->MLbase))
                            * (Tight+QDT[12][indxR[jr]+k1+1][isjs]);
                    }
                    for (k2=is; k2<js-1; k2++) {
                        G = 0;
                        for (k=1; k<=num_align; k++)
                            G += P->MLintern[pair_typeR[k][irjr]];
                        
                        Tight = QTu[3][irjr][indxS[js]+k2+1] * EXP(((double) G)/num_align);
                        QRT[12][irjr][isjs] += (QfS[indxS[k2]+is]+EXP((k2-is+1)*BETA3))
                            * (Tight+QDT[12][irjr][indxS[js]+k2+1]);
                    }
                    //fk
                    Tight = QTu[3][irjr][isjs]*EXP(BETA2);
                    QRT[13][irjr][isjs] = Tight + QDT[13][irjr][isjs];
                    for (k1=ir; k1<jr-1; k1++)
                        for (k2=is; k2<js-1; k2++) {
                            Tight = QTu[3][indxR[jr]+k1+1][indxS[js]+k2+1] * EXP(BETA2);
                            QRT[13][irjr][isjs] += (QfR[indxR[k1]+ir]+EXP((k1-ir+1)*BETA3))
                                * (QfS[indxS[k2]+is]+EXP((k2-is+1)*BETA3))
                                * (Tight+QDT[13][indxR[jr]+k1+1][indxS[js]+k2+1]);
                        }
                    for (k1=ir; k1<jr-1; k1++) {
                        Tight = QTu[3][indxR[jr]+k1+1][isjs] * EXP(BETA2);
                        QRT[13][irjr][isjs] += (QfR[indxR[k1]+ir]+EXP((k1-ir+1)*BETA3))
                            * (Tight+QDT[13][indxR[jr]+k1+1][isjs]);
                    }
                    for (k2=is; k2<js-1; k2++) {
                        Tight = QTu[3][irjr][indxS[js]+k2+1] * EXP(BETA2);
                        QRT[13][irjr][isjs] += (QfS[indxS[k2]+is]+EXP((k2-is+1)*BETA3))
                            * (Tight+QDT[13][irjr][indxS[js]+k2+1]);
                    }
                    //kea
                    Tight = QTd[3][irjr][isjs];
                    QRT[14][irjr][isjs] = Tight + QDT[14][irjr][isjs];
                    for (k1=ir; k1<jr-1; k1++)
                        for (k2=is; k2<js-1; k2++) {
                            Tight = QTd[3][indxR[jr]+k1+1][indxS[js]+k2+1];
                            QRT[14][irjr][isjs] += (QfR[indxR[k1]+ir]+EXP((k1-ir+1)*BETA3))
                                * QS[indxS[k2]+is] * (Tight+QDT[14][indxR[jr]+k1+1][indxS[js]+k2+1]);
                            QRT[14][irjr][isjs] += (QfR[indxR[k1]+ir]*QS[indxS[k2]+is]
                                                    +EXP((k1-ir+1)*BETA3)*(QS[indxS[k2]+is]-1))
                                * QDT[15][indxR[jr]+k1+1][indxS[js]+k2+1];
                        }
                    for (k1=ir; k1<jr-1; k1++) {
                        Tight = QTd[3][indxR[jr]+k1+1][isjs];
                        QRT[14][irjr][isjs] += (QfR[indxR[k1]+ir]+EXP((k1-ir+1)*BETA3))
                            * (Tight+QDT[14][indxR[jr]+k1+1][isjs]);
                        QRT[14][irjr][isjs] += QfR[indxR[k1]+ir] * QDT[15][indxR[jr]+k1+1][isjs];
                    }
                    for (k2=is; k2<js-1; k2++) {
                        Tight = QTd[3][irjr][indxS[js]+k2+1];
                        QRT[14][irjr][isjs] += QS[indxS[k2]+is] * (Tight+QDT[14][irjr][indxS[js]+k2+1]);
                        QRT[14][irjr][isjs] += (QS[indxS[k2]+is]-1) * QDT[15][irjr][indxS[js]+k2+1];
                    }
                    //keb
                    for (k1=ir; k1<jr; k1++)
                        for (k2=is; k2<js; k2++)
                            QRT[15][irjr][isjs] += QDT[15][indxR[jr]+k1][indxS[js]+k2]
                                * EXP((k1-ir)*BETA3);
                    //km
                    G = 0;
                    for (k=1; k<=num_align; k++)
                        G += P->MLintern[rtype[pair_typeS[k][isjs]]];
                    
                    Tight = QTd[3][irjr][isjs] * EXP(((double) G)/num_align);
                    QRT[16][irjr][isjs] = Tight + QDT[16][irjr][isjs];
                    for (k1=ir; k1<jr-1; k1++)
                        for (k2=is; k2<js-1; k2++) {
                            GMS = 0;
                            for (k=1; k<=num_align; k++)
                                GMS += P->MLintern[rtype[pair_typeS[k][indxS[js]+k2+1]]];
                            
                            Tight = QTd[3][indxR[jr]+k1+1][indxS[js]+k2+1] * EXP(((double) GMS)/num_align);
                            QRT[16][irjr][isjs] += (QfR[indxR[k1]+ir]+EXP((k1-ir+1)*BETA3))
                                * (QmS[indxS[k2]+is]+EXP((k2-is+1)*P->MLbase))
                                * (Tight+QDT[16][indxR[jr]+k1+1][indxS[js]+k2+1]);
                        }
                    for (k1=ir; k1<jr-1; k1++) {
                        G = 0;
                        for (k=1; k<=num_align; k++)
                            G += P->MLintern[rtype[pair_typeS[k][isjs]]];
                        
                        Tight = QTd[3][indxR[jr]+k1+1][isjs] * EXP(((double) G)/num_align);
                        QRT[16][irjr][isjs] += (QfR[indxR[k1]+ir]+EXP((k1-ir+1)*BETA3))
                            * (Tight+QDT[16][indxR[jr]+k1+1][isjs]);
                    }
                    for (k2=is; k2<js-1; k2++) {
                        GMS = 0;
                        for (k=1; k<=num_align; k++)
                            GMS += P->MLintern[rtype[pair_typeS[k][indxS[js]+k2+1]]];
                        
                        Tight = QTd[3][irjr][indxS[js]+k2+1] * EXP(((double) GMS)/num_align);
                        QRT[16][irjr][isjs] += (QmS[indxS[k2]+is]+EXP((k2-is+1)*P->MLbase))
                            * (Tight+QDT[16][irjr][indxS[js]+k2+1]);
                    }
                    //kf
                    Tight = QTd[3][irjr][isjs] * EXP(BETA2);
                    QRT[17][irjr][isjs] = Tight + QDT[17][irjr][isjs];
                    for (k1=ir; k1<jr-1; k1++)
                        for (k2=is; k2<js-1; k2++) {
                            Tight = QTd[3][indxR[jr]+k1+1][indxS[js]+k2+1] * EXP(BETA2);
                            QRT[17][irjr][isjs] += (QfR[indxR[k1]+ir]+EXP((k1-ir+1)*BETA3))
                                * (QfS[indxS[k2]+is]+EXP((k2-is+1)*BETA3))
                                * (Tight+QDT[17][indxR[jr]+k1+1][indxS[js]+k2+1]);
                        }
                    for (k1=ir; k1<jr-1; k1++) {
                        Tight = QTd[3][indxR[jr]+k1+1][isjs] * EXP(BETA2);
                        QRT[17][irjr][isjs] += (QfR[indxR[k1]+ir]+EXP((k1-ir+1)*BETA3))
                            * (Tight+QDT[17][indxR[jr]+k1+1][isjs]);
                    }
                    for (k2=is; k2<js-1; k2++) {
                        Tight = QTd[3][irjr][indxS[js]+k2+1] * EXP(BETA2);
                        QRT[17][irjr][isjs] += (QfS[indxS[k2]+is]+EXP((k2-is+1)*BETA3))
                            * (Tight+QDT[17][irjr][indxS[js]+k2+1]);
                    }
                    //kka
                    QRT[18][irjr][isjs] = QDT[18][irjr][isjs];
                    for (k1=ir; k1<jr-1; k1++)
                        for (k2=is; k2<js-1; k2++) {
                            QRT[18][irjr][isjs] += (QfR[indxR[k1]+ir]+EXP((k1-ir+1)*BETA3))
                                * (QfS[indxS[k2]+is]+EXP((k2-is+1)*BETA3))
                                * (QDT[18][indxR[jr]+k1+1][indxS[js]+k2+1]);
                            QRT[18][irjr][isjs] += (QfR[indxR[k1]+ir]*QfS[indxS[k2]+is]
                                                    +EXP((k1-ir+1)*BETA3)*QfS[indxS[k2]+is]
                                                    +QfR[indxR[k1]+ir]*EXP((k2-is+1)*BETA3))
                                * QDT[19][indxR[jr]+k1+1][indxS[js]+k2+1];
                        }
                    for (k1=ir; k1<jr-1; k1++) {
                        QRT[18][irjr][isjs] += (QfR[indxR[k1]+ir]+EXP((k1-ir+1)*BETA3))
                            * QDT[18][indxR[jr]+k1+1][isjs];
                        QRT[18][irjr][isjs] += QfR[indxR[k1]+ir] * QDT[19][indxR[jr]+k1+1][isjs];
                    }
                    for (k2=is; k2<js-1; k2++) {
                        QRT[18][irjr][isjs] += (QfS[indxS[k2]+is]+EXP((k2-is+1)*BETA3))
                            * QDT[18][irjr][indxS[js]+k2+1];
                        QRT[18][irjr][isjs] += QfS[indxS[k2]+is] * QDT[19][irjr][indxS[js]+k2+1];
                    }
                    //kkb
                    for (k1=ir; k1<jr; k1++)
                        for (k2=is; k2<js; k2++) {
                            QRT[19][irjr][isjs] += QDT[19][indxR[jr]+k1][indxS[js]+k2]
                                * EXP((k1-ir)*BETA3) * EXP((k2-is)*BETA3);
                        }
                }

    //partition function of arbitrary RIP structure
	Qany = QR[indxR[len_r]+1]*QS[indxS[len_s]+1]
        + (QRT[0][indxR[len_r]+1][indxS[len_s]+1]+QRT[1][indxR[len_r]+1][indxS[len_s]+1]);
	for (k1=2; k1<len_r; k1++) 
        for (k2=2; k2<len_s; k2++)
            Qany += (QRT[0][indxR[k1]+1][indxS[k2]+1]+QRT[1][indxR[k1]+1][indxS[k2]+1])
                * QR[indxR[len_r]+k1+1] * QS[indxS[len_s]+k2+1];
    for (k1=2; k1<len_r; k1++)
        Qany += (QRT[0][indxR[k1]+1][indxS[len_s]+1]+QRT[1][indxR[k1]+1][indxS[len_s]+1])
            * QR[indxR[len_r]+k1+1];
	for (k2=2; k2<len_s; k2++)
        Qany += (QRT[0][indxR[len_r]+1][indxS[k2]+1]+QRT[1][indxR[len_r]+1][indxS[k2]+1])
            * QS[indxS[len_s]+k2+1];
    /* end of calculating partition function */

    fprintf(fp, "PF of RIP: (%d %d %d %d)  %16.16e\n", 1, len_r, 1, len_s, Qany);
    
    #ifndef     DEBUG
    void print_align_pfunc();
    print_align_pfunc();
    #endif
    
	free(seqs_s_35);
}

void print_align_pfunc()
{
    int ir, jr, is, js;
    int irjr, isjs;
    int gr, gs;

    for (gr=1; gr<len_r; gr++)
        for (ir=1; ir<=len_r-gr; ir++)
            for (gs=1; gs<len_s; gs++)
                for (is=1; is<=len_s-gs; is++) {
                    jr = ir+gr;
                    js = is+gs;
                    irjr = indxR[jr]+ir;
                    isjs = indxS[js]+is;

                    /*
                    #ifdef  TEST                    
                    //Hybrid
                    if (Bond[0][irjr][isjs] != 0)
                        printf("%2d %2d %2d %2d:  %16.16e\n", ir, jr, is, js, Bond[0][irjr][isjs]);
                    #endif
                    */

                    /*
                    #ifndef  TEST
                    //Tsu
                    if (QTsu[0][irjr][isjs] != 0)
                        printf("%2d %2d %2d %2d:  %16.16e\n", ir, jr, is, js, QTsu[0][irjr][isjs]);
                    #else
                    if (QTsd[0][irjr][isjs] != 0)
                        printf("%2d %2d %2d %2d:  %16.16e\n", ir, jr, is, js, QTsd[0][irjr][isjs]);
                    #endif
                    */

                    /*
                    #ifdef  TEST
                    //Tu
                    if (QTu[0][irjr][isjs] != 0)
                        printf("%2d %2d %2d %2d:  %16.16e\n", ir, jr, is, js, QTu[0][irjr][isjs]);
                    #else
                    if (QTd[0][irjr][isjs] != 0)
                        printf("%2d %2d %2d %2d:  %16.16e\n", ir, jr, is, js, QTd[0][irjr][isjs]);
                    #endif
                    */

                    /*
                    #ifndef  TEST
                    //Tsd
                    if (QTsd[3][irjr][isjs] != 0)
                        printf("%2d %2d %2d %2d:  %16.16e\n", ir,jr,is,js,QTsd[3][irjr][isjs]);
                    #else
                    if (QTsu[3][irjr][isjs] != 0)
                        printf("%2d %2d %2d %2d:  %16.16e\n", ir,jr,is,js,QTsu[3][irjr][isjs]);
                    #endif
                    */

                    /*
                    #ifndef  TEST
                    //Td
                    if (QTd[0][irjr][isjs]!=0)
                        printf("%2d %2d %2d %2d:  %16.16e\n", ir,jr,is,js,QTd[0][irjr][isjs]);
                    #else
                    if (QTu[0][irjr][isjs]!=0)
                        printf("%2d %2d %2d %2d:  %16.16e\n", ir,jr,is,js,QTu[0][irjr][isjs]);
                    #endif
                    */

                    /*
                    #ifdef  TEST
                    //Te
                    if (QTe[irjr][isjs] != 0)
                        printf("%2d %2d %2d %2d:  %16.16e\n", ir, jr, is, js, QTe[irjr][isjs]);
                    #endif
                    */

                    /*
                    #ifndef  TEST
                    //DT
                    if (QDT[2][irjr][isjs] != 0)
                        printf("%2d %2d %2d %2d:  %16.16e\n", ir, jr, is, js, QDT[2][irjr][isjs]);
                    #else
                    if (QDT[3][irjr][isjs] != 0)
                        printf("%2d %2d %2d %2d:  %16.16e\n", ir, jr, is, js, QDT[3][irjr][isjs]);
                    #endif
                    */

                    /*
                    #ifdef  TEST
                    //DT
                    if (QDT[9][irjr][isjs] != 0)
                        printf("%2d %2d %2d %2d:  %16.16e\n", ir, jr, is, js, QDT[9][irjr][isjs]);
                    #endif
                    */

                    
                    #ifdef  TEST
                    if (QDT[10][irjr][isjs]!=0 || QDT[11][irjr][isjs]!=0)
                        printf("%2d %2d %2d %2d:  %16.16e\n",  ir, jr, is, js,
                               QDT[10][irjr][isjs]+QDT[11][irjr][isjs]);
                    #else
                    if (QDT[14][irjr][isjs]!=0 || QDT[15][irjr][isjs]!=0)
                        printf("%2d %2d %2d %2d:  %16.16e\n",  ir, jr, is, js,
                               QDT[14][irjr][isjs]+QDT[15][irjr][isjs]);
                    #endif
                    

                    /*
                    #ifdef  TEST
                    if (QDT[18][irjr][isjs]!=0 || QDT[19][irjr][isjs]!=0)
                        printf("%2d %2d %2d %2d:  %16.16e\n",  ir, jr, is, js,
                               QDT[18][irjr][isjs]+QDT[19][irjr][isjs]);
                    #endif
                    */
                }
}
